from __future__ import annotations

from collections import Counter
from collections.abc import Iterator
from itertools import groupby

from DAJIN2.core.clustering.kmer_generator import generate_mutation_kmers


def subset_scores(labels: list[int], scores: list[int], label_most: int, length: int = 1000) -> list[int]:
    subset = [score for label, score in zip(labels, scores) if label == label_most]
    return subset[:length]


def call_count(midsv_tags: Iterator[list[str]]) -> list[dict[str, int]]:
    return [dict(Counter(midsv)) for midsv in zip(*midsv_tags)]


def call_percent(counts: list[dict[str, int]]) -> list[dict[str, float]]:
    coverage = sum(counts[0].values())
    return [{k: round(v / coverage * 100, 3) for k, v in count.items()} for count in counts]


def subtract_percentage(
    percent_sample: list[dict[str, float]], percent_control: list[dict[str, float]], knockin_loci: set[int]
) -> list[dict[str, float]]:
    """Note: The subtraction operation with collections.Counter class removes elements from the result that have a value of 0 or negative."""
    return [
        dict(samp) if i in knockin_loci else dict(Counter(samp) - Counter(cont))
        for i, (samp, cont) in enumerate(zip(percent_sample, percent_control))
    ]


def discard_common_error(percent_subtraction: list[dict[str, float]], threshold: float = 0.5) -> list[dict]:
    return [{k: v for k, v in samp.items() if v > threshold} for samp in percent_subtraction]


def discard_matches_and_ns(percent_discarded: list[dict[str, float]]) -> list[dict[str, float]]:
    result = []
    for mutation_percent in percent_discarded:
        filtered_dict = {}
        for mutation, percent in mutation_percent.items():
            # Extract the center part of the mutation string
            mutation_center = mutation.split(",")[1]
            # Skip mutations that start with '='
            if mutation_center.startswith("="):
                continue
            # Add to the filtered dictionary if it passed the conditions
            filtered_dict[mutation] = percent
        result.append(filtered_dict)

    return result


###############################################################################
# Handling insertions
###############################################################################


def group_consecutive_insertions(mutation_loci: dict[str, set[int]]) -> list[tuple[int]]:
    index = sorted(i for i, m in enumerate(mutation_loci) if "+" in m)
    index_grouped = []
    for _, group in groupby(enumerate(index), lambda i_x: i_x[0] - i_x[1]):
        items = [value for _, value in group]
        index_grouped.append(tuple(items))
    return index_grouped


def find_max_insertion_score_in_group(percent_discarded: list[dict[str, float]], index_group: tuple[int]) -> float:
    max_score = 0
    for i in index_group:
        for key, val in percent_discarded[i].items():
            if key.split(",")[1].startswith("+"):
                max_score = max(max_score, val)
    return max_score


def update_insertion_scores_in_group_to_max(
    percent_discarded: list[dict[str, float]], index_group: tuple[int], max_score: float
) -> list[dict[str, float]]:
    for i in index_group:
        for key in percent_discarded[i]:
            if key.split(",")[1].startswith("+"):
                percent_discarded[i][key] = max_score
    return percent_discarded


def update_insertion_score(
    percent_discarded: list[dict[str, float]], mutation_loci: dict[str, set[int]]
) -> list[dict[str, float]]:
    """Update the insertion scores in the percent_discarded list based on the max score in each group of consecutive insertions.

    This function groups consecutive insertion indices and updates the scores of insertions within those groups to the maximum score found in each group.

    Args:
        percent_discarded (list[dict[str, float]]): A list of dictionaries. Each dictionary contains the scores of mutations at a particular locus, represented as a percentage.
        mutation_loci (dict[str, set[int]]): A dictionary mapping mutation types to sets of indices where those mutations occur.

    Returns:
        list[dict[str, float]]: The updated list of dictionaries containing the new insertion scores.

    """
    index_grouped = group_consecutive_insertions(mutation_loci)
    for index_insertion in index_grouped:
        max_score = find_max_insertion_score_in_group(percent_discarded, index_insertion)
        percent_discarded = update_insertion_scores_in_group_to_max(percent_discarded, index_insertion, max_score)
    return percent_discarded


###############################################################################
# make_score
###############################################################################


def make_score(
    path_sample, path_control, mutation_loci: list[set[int]], knockin_loci: list[set[int]]
) -> list[dict[str, float]]:
    midsv_tags_sample = generate_mutation_kmers(path_sample, mutation_loci, compress_ins=True)
    midsv_tags_control = generate_mutation_kmers(path_control, mutation_loci, compress_ins=True)
    counts_sample = call_count(midsv_tags_sample)
    counts_control = call_count(midsv_tags_control)
    percent_sample = call_percent(counts_sample)
    percent_control = call_percent(counts_control)
    percent_subtraction = subtract_percentage(percent_sample, percent_control, knockin_loci)
    percent_discarded = discard_common_error(percent_subtraction, 0.5)
    percent_discarded = update_insertion_score(percent_discarded, mutation_loci)
    mutation_score = discard_matches_and_ns(percent_discarded)
    return mutation_score


###############################################################################
# annotate_score
###############################################################################


def annotate_score(path_sample, mutation_score, mutation_loci, is_control=False) -> Iterator[list[float]]:
    for midsv_kmer in generate_mutation_kmers(path_sample, mutation_loci):
        score = [0 for _ in range(len(midsv_kmer))]
        for i, (cs_kmer, mut_score) in enumerate(zip(midsv_kmer, mutation_score)):
            if mut_score == {}:
                continue
            # Mutation sites are not considered in controls because they should be sample-specific.
            if is_control and cs_kmer.split(",")[1][0] in mutation_loci[i]:
                continue
            if cs_kmer in mut_score:
                score[i] = mut_score[cs_kmer]
        yield score



from __future__ import annotations

import logging
import shutil
from pathlib import Path

from DAJIN2.core import classification, clustering, consensus, preprocess, report
from DAJIN2.core.preprocess.infrastructure.input_formatter import FormattedInputs
from DAJIN2.utils import allele_handler, fastx_handler, fileio

logger = logging.getLogger(__name__)


###########################################################
# main
###########################################################


def execute_control(arguments: dict):
    logger.info(f"{arguments['control']} is now processing...")

    ###########################################################
    # Setup arguments
    ###########################################################
    ARGS: FormattedInputs = preprocess.format_inputs(arguments)
    preprocess.create_temporal_directories(ARGS.tempdir, ARGS.control_name, is_control=True)
    preprocess.create_report_directories(ARGS.tempdir, ARGS.control_name, is_control=True)
    allele_handler.save_allele_name_map(ARGS.tempdir, ARGS.fasta_alleles.keys())
    fileio.cache_file_hash(ARGS.path_allele, Path(ARGS.tempdir, "cache", "hash_allele.txt"))

    ###########################################################
    # Check caches
    ###########################################################
    if Path(ARGS.tempdir, "report", "BAM", ARGS.control_name, f"{ARGS.control_name}.bam").exists():
        logger.info(f"{arguments['control']} is already preprocessed and reuse the results for the current run...")
        return None

    logger.info(f"Preprocess {arguments['control']}...")

    ###########################################################
    # Concatenate fastq files
    ###########################################################

    fastx_handler.save_inputs_as_single_fastq(ARGS, is_control=True)

    path_fastq = Path(ARGS.tempdir, ARGS.control_name, "fastq", f"{ARGS.control_name}.fastq.gz")
    fastx_handler.overwrite_with_downsampled_fastq(path_fastq, num_reads=20_000)

    ###########################################################
    # Export fasta files as single-FASTA format
    ###########################################################
    fastx_handler.export_fasta_files(ARGS, is_control=True)

    ###########################################################
    # Separate fastq files by sequence error
    ###########################################################
    control_allele_key = allele_handler.to_allele_key("control")
    paths_fasta = [Path(ARGS.tempdir, ARGS.control_name, "fasta", f"{control_allele_key}.fasta")]
    preprocess.generate_sam(ARGS, paths_fasta, is_control=True, is_sv=False)
    preprocess.generate_midsv(ARGS, is_control=True, is_sv=False)
    # ============================================================
    # Detect sequence error reads
    # ============================================================
    preprocess.detect_sequence_error_reads(ARGS, is_control=True)
    preprocess.split_fastq_by_sequence_error(ARGS, is_control=True)

    # ============================================================
    # Save subsetted fastq if the read number is too large (> 10,000 reads)
    # ============================================================
    path_fastq = Path(ARGS.tempdir, ARGS.control_name, "fastq", f"{ARGS.control_name}.fastq.gz")
    path_fastq_with_error = Path(
        ARGS.tempdir, ARGS.control_name, "fastq", f"{ARGS.control_name}_sequence_error.fastq.gz"
    )
    fastx_handler.overwrite_with_downsampled_fastq(path_fastq, num_reads=10_000)
    fastx_handler.overwrite_with_downsampled_fastq(path_fastq_with_error, num_reads=10_000)

    ###########################################################
    # Mapping filtered control reads
    ###########################################################

    # ============================================================
    # Mapping using mappy
    # ============================================================
    paths_fasta = Path(ARGS.tempdir, ARGS.control_name, "fasta").glob("*.fasta")
    preprocess.generate_sam(ARGS, paths_fasta, is_control=True, is_sv=False)

    # ============================================================
    # MIDSV conversion
    # ============================================================
    preprocess.generate_midsv(ARGS, is_control=True, is_sv=False)

    ###########################################################
    # Prepare data to `extract mutaion loci`
    ###########################################################
    preprocess.cache_mutation_loci(ARGS, is_control=True)

    ###########################################################
    # Output BAM files
    ###########################################################
    logger.info(f"Output BAM files of {arguments['control']}...")
    report.bam_exporter.export_to_bam(
        ARGS.tempdir, ARGS.control_name, ARGS.genome_coordinates, ARGS.threads, is_control=True
    )
    ###########################################################
    # Finish call
    ###########################################################
    logger.info(f"\N{TEACUP WITHOUT HANDLE} {arguments['control']} is finished!")


def execute_sample(arguments: dict):
    logger.info(f"{arguments['sample']} is now processing...")

    ###########################################################
    # Preprocess
    ###########################################################

    ARGS: FormattedInputs = preprocess.format_inputs(arguments)
    preprocess.create_temporal_directories(ARGS.tempdir, ARGS.sample_name, is_control=False)
    preprocess.create_report_directories(ARGS.tempdir, ARGS.sample_name, is_control=False)
    allele_handler.save_allele_name_map(ARGS.tempdir, ARGS.fasta_alleles.keys())

    logger.info(f"Preprocess {arguments['sample']}...")

    # ============================================================
    # Merge fastq files
    # ============================================================

    fastx_handler.save_inputs_as_single_fastq(ARGS, is_control=False)

    # Save subsetted fastq if the read number is too large (> 100,000 reads)
    path_fastq = Path(ARGS.tempdir, ARGS.sample_name, "fastq", f"{ARGS.sample_name}.fastq.gz")
    fastx_handler.overwrite_with_downsampled_fastq(path_fastq, num_reads=100_000)

    # ============================================================
    # Mapping with mappy
    # ============================================================

    for path_fasta in Path(ARGS.tempdir, ARGS.control_name, "fasta").glob("*.fasta"):
        shutil.copy(path_fasta, Path(ARGS.tempdir, ARGS.sample_name, "fasta"))

    paths_fasta = Path(ARGS.tempdir, ARGS.sample_name, "fasta").glob("*.fasta")
    preprocess.generate_sam(ARGS, paths_fasta, is_control=False, is_sv=False)

    # ============================================================
    # MIDSV conversion
    # ============================================================

    preprocess.generate_midsv(ARGS, is_control=False, is_sv=False)

    ###########################################################
    # Remove sequence error reads from MIDSV files
    ###########################################################
    preprocess.detect_sequence_error_reads(ARGS, is_control=False)
    preprocess.split_fastq_by_sequence_error(ARGS, is_control=False)
    preprocess.replace_midsv_without_sequence_errors(ARGS)

    # ============================================================
    # Extract mutation loci
    # ============================================================
    preprocess.extract_knockin_loci(ARGS.tempdir, ARGS.sample_name)
    preprocess.cache_mutation_loci(ARGS, is_control=False)

    # ============================================================
    # Detect and mapping SV alleles
    # ============================================================
    paths_predefined_fasta: set[str] = {
        str(
            Path(
                ARGS.tempdir,
                ARGS.sample_name,
                "fasta",
                f"{allele_handler.to_allele_key(allele)}.fasta",
            )
        )
        for allele in ARGS.fasta_alleles.keys()
    }

    preprocess.detect_sv_alleles(
        ARGS.tempdir, ARGS.sample_name, ARGS.control_name, ARGS.fasta_alleles, sv_type="insertion"
    )
    preprocess.detect_sv_alleles(
        ARGS.tempdir, ARGS.sample_name, ARGS.control_name, ARGS.fasta_alleles, sv_type="deletion"
    )
    preprocess.detect_sv_alleles(
        ARGS.tempdir, ARGS.sample_name, ARGS.control_name, ARGS.fasta_alleles, sv_type="inversion"
    )

    paths_all_fasta = {str(p) for p in Path(ARGS.tempdir, ARGS.sample_name, "fasta").glob("*.fasta")}
    paths_sv_fasta = paths_all_fasta - paths_predefined_fasta

    if paths_sv_fasta:
        # mapping to SV alleles
        preprocess.generate_sam(ARGS, paths_sv_fasta, is_control=True, is_sv=True)
        preprocess.generate_sam(ARGS, paths_sv_fasta, is_control=False, is_sv=True)
        # add SV alleles to ARGS.fasta_alleles
        for path_fasta_sv in paths_sv_fasta:
            allele, seq = Path(path_fasta_sv).read_text().strip().split("\n")
            allele = allele.replace(">", "")
            ARGS.fasta_alleles[allele] = seq
        # MIDSV conversion
        preprocess.generate_midsv(ARGS, is_control=True, is_sv=True)
        preprocess.generate_midsv(ARGS, is_control=False, is_sv=True)
        # Reculculate mutation loci
        preprocess.cache_mutation_loci(ARGS, is_control=True)
        preprocess.extract_knockin_loci(ARGS.tempdir, ARGS.sample_name)
        preprocess.cache_mutation_loci(ARGS, is_control=False)

    allele_handler.save_allele_name_map(ARGS.tempdir, ARGS.fasta_alleles.keys())
    fileio.save_pickle(ARGS.fasta_alleles, Path(ARGS.tempdir, ARGS.sample_name, "fasta", "fasta_alleles.pickle"))

    ########################################################################
    # Classify alleles
    ########################################################################

    logger.info(f"Classify {arguments['sample']}...")

    classif_sample = classification.classify_alleles(
        ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name, ARGS.no_filter
    )

    fileio.save_pickle(classif_sample, Path(ARGS.tempdir, ARGS.sample_name, "classification", "classif_sample.pickle"))

    ########################################################################
    # Clustering
    ########################################################################

    logger.info(f"Clustering {arguments['sample']}...")

    labels = clustering.extract_labels(classif_sample, ARGS.tempdir, ARGS.sample_name, ARGS.control_name)
    clust_sample = clustering.add_labels(classif_sample, labels)
    clust_sample = clustering.add_readnum(clust_sample)
    clust_sample = clustering.add_percent(clust_sample)
    clust_sample = clustering.update_labels(clust_sample)

    fileio.save_pickle(clust_sample, Path(ARGS.tempdir, ARGS.sample_name, "clustering", "clust_sample.pickle"))

    ########################################################################
    # Consensus call
    ########################################################################

    logger.info(f"Consensus calling of {arguments['sample']}...")

    # Remove minor alleles with fewer than 5 reads or less than 0.5%
    clust_sample_removed = consensus.remove_minor_alleles(clust_sample, ARGS.no_filter)

    # Adjust the percentage to 100%
    clust_sample_removed = consensus.scale_percentage(clust_sample_removed)

    # Downsampling to 1000 reads in each LABEL
    clust_downsampled = consensus.downsample_by_label(clust_sample_removed, 1000)

    consensus.cache_mutation_loci(ARGS, clust_downsampled, ARGS.no_filter)

    cons_percentages, cons_sequences, cons_midsv_tags, label_before_to_after = consensus.call_consensus(
        ARGS.tempdir, ARGS.sample_name, ARGS.fasta_alleles, clust_downsampled
    )

    map_label_name, map_name_allele = consensus.call_allele_name(cons_sequences, ARGS.fasta_alleles)
    cons_percentages = consensus.update_key_by_allele_name(cons_percentages, map_label_name)
    cons_sequences = consensus.update_key_by_allele_name(cons_sequences, map_label_name)
    cons_midsv_tags = consensus.update_key_by_allele_name(cons_midsv_tags, map_label_name)

    fileio.save_pickle(cons_percentages, Path(ARGS.tempdir, ARGS.sample_name, "consensus", "cons_percentages.pickle"))
    fileio.save_pickle(cons_sequences, Path(ARGS.tempdir, ARGS.sample_name, "consensus", "cons_sequences.pickle"))
    fileio.save_pickle(cons_midsv_tags, Path(ARGS.tempdir, ARGS.sample_name, "consensus", "cons_midsv_tags.pickle"))
    fileio.save_pickle(map_label_name, Path(ARGS.tempdir, ARGS.sample_name, "consensus", "map_label_name.pickle"))
    fileio.save_pickle(map_name_allele, Path(ARGS.tempdir, ARGS.sample_name, "consensus", "map_name_allele.pickle"))

    ########################################################################
    # Output Report：RESULT/FASTA/HTML/BAM
    ########################################################################

    logger.info(f"Output reports of {arguments['sample']}...")

    # RESULT
    RESULT_SAMPLE = consensus.update_label_percent_readnum_name(
        clust_sample_removed, map_label_name, label_before_to_after
    )
    RESULT_SAMPLE.sort(key=lambda x: x["LABEL"])

    fileio.write_jsonl(RESULT_SAMPLE, Path(ARGS.tempdir, "result", f"{ARGS.sample_name}.jsonl"))

    # FASTA
    report.sequence_exporter.export_to_fasta(ARGS.tempdir, ARGS.sample_name, cons_sequences)
    report.sequence_exporter.export_reference_to_fasta(ARGS.tempdir, ARGS.sample_name)

    # HTML
    report.sequence_exporter.export_to_html(
        ARGS.tempdir, ARGS.sample_name, ARGS.fasta_alleles, cons_midsv_tags, map_name_allele
    )

    # VCF
    report.vcf_exporter.export_to_vcf(ARGS.tempdir, ARGS.sample_name, ARGS.genome_coordinates, cons_midsv_tags)

    # BAM
    report.bam_exporter.export_to_bam(
        ARGS.tempdir, ARGS.sample_name, ARGS.genome_coordinates, ARGS.threads, RESULT_SAMPLE
    )
    for path_bam_igvjs in Path(ARGS.tempdir, "cache", ".igvjs").glob(f"{ARGS.control_name}_control.bam*"):
        shutil.copy(path_bam_igvjs, Path(ARGS.tempdir, "report", ".igvjs", ARGS.sample_name))

    ###########################################################
    # Finish call
    ###########################################################

    logger.info(f"\N{TEACUP WITHOUT HANDLE} {arguments['sample']} is finished!")



from __future__ import annotations

import datetime
import logging
import sys
import warnings
from dataclasses import dataclass
from pathlib import Path

from sklearn.exceptions import ConvergenceWarning

DAJIN_RESULTS_DIR = Path("DAJIN_Results")
TEMP_ROOT_DIR = Path(DAJIN_RESULTS_DIR, ".tempdir")


@dataclass(frozen=True)
class ConsensusKey:
    allele: str
    label: int
    percent: float


class DeferredFileHandler(logging.FileHandler):
    def __init__(self, filename, mode="a", encoding=None, delay=True):
        # Setting delay to True to defer the file opening
        super().__init__(filename, mode, encoding, delay)

    def emit(self, record):
        # The file is actually opened when a log is emitted
        if self.stream is None:
            self.stream = self._open()
        super().emit(record)


def get_logfile() -> Path:
    current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S_%f")
    return Path(f"DAJIN2_log_{current_time}.txt")


def set_logging(path_logfile: Path, level: int = logging.INFO) -> logging.Logger:
    log_format = "%(asctime)s, %(levelname)s, %(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"

    stderr_handler = logging.StreamHandler()
    stderr_handler.setFormatter(logging.Formatter(log_format, datefmt=date_format))

    file_handler = DeferredFileHandler(path_logfile)
    file_handler.setFormatter(logging.Formatter(log_format, datefmt=date_format))

    logging.basicConfig(level=level, handlers=[stderr_handler, file_handler])

    # Drop noisy kaleido Chromium launch message (see issue #117)
    def _suppress_chromium_init(record: logging.LogRecord) -> bool:
        return "Chromium init'ed with kwargs" not in record.getMessage()

    chromium_filter = logging.Filter()
    chromium_filter.filter = _suppress_chromium_init
    stderr_handler.addFilter(chromium_filter)
    file_handler.addFilter(chromium_filter)

    # log uncaught exceptions
    def handle_uncaught_exception(exc_type, exc_value, exc_traceback):
        logging.error("Catch an Exception. Traceback:", exc_info=(exc_type, exc_value, exc_traceback))

    sys.excepthook = handle_uncaught_exception

    return logging.getLogger(__name__)


def reset_logging() -> None:
    """
    Reset the logging system to its default state.
    """
    # Remove all existing handlers
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)


def set_warnings_ignore() -> None:
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    warnings.filterwarnings("ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=ConvergenceWarning)



from __future__ import annotations

import pickle
import tempfile
from pathlib import Path

import midsv

from DAJIN2.core.preprocess.alignment import mapping
from DAJIN2.utils.allele_handler import to_allele_key

###########################################################
# Consider all mutations are possible in the knockin region
# For large deletion alleles, the deleted sequence becomes the knock-in region, so all mutations within this region are taken into consideration.
# The code identifies the flox knock-in sites as deletions not present in the control.
###########################################################


def is_valid_file(control_fasta: Path, knockin_fasta: Path) -> bool:
    return knockin_fasta != control_fasta and knockin_fasta.suffix == ".fasta"


def get_index_of_knockin_loci(control_fasta: Path, knockin_fasta: Path) -> set[int]:
    alignments = mapping.to_sam(knockin_fasta, control_fasta, preset="map-ont")
    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".sam") as tmp:
        path_sam = Path(tmp.name)
        for alignment in alignments:
            tmp.write(alignment if alignment.endswith("\n") else f"{alignment}\n")

    try:
        alignments_midsv = midsv.transform(path_sam=path_sam, qscore=False)
        if not alignments_midsv:
            return set()
        midsv_tags = alignments_midsv[0]["MIDSV"].split(",")
        return {i for i, cs in enumerate(midsv_tags) if cs.startswith("-")}
    finally:
        path_sam.unlink(missing_ok=True)


def extract_knockin_loci(TEMPDIR: str | Path, SAMPLE_NAME: str) -> None:
    control_fasta = Path(TEMPDIR, SAMPLE_NAME, "fasta", f"{to_allele_key('control')}.fasta")
    for knockin_fasta in Path(TEMPDIR, SAMPLE_NAME, "fasta").iterdir():
        if not is_valid_file(control_fasta, knockin_fasta):
            continue

        allele_key = knockin_fasta.stem
        path_output = Path(TEMPDIR, SAMPLE_NAME, "knockin_loci", allele_key, "knockin.pickle")
        path_output.parent.mkdir(parents=True, exist_ok=True)
        if path_output.exists():
            continue

        knockin_loci = get_index_of_knockin_loci(control_fasta, knockin_fasta)

        with open(path_output, "wb") as p:
            pickle.dump(knockin_loci, p)



from DAJIN2.core.clustering.appender import add_labels, add_percent, add_readnum
from DAJIN2.core.clustering.label_extractor import extract_labels
from DAJIN2.core.clustering.label_updator import update_labels



from __future__ import annotations


def relabel_with_consective_order(labels: list[int], start: int = 1) -> list[int]:
    """
    Reorders the given list of labels starting from the given start value.

    Args:
    - labels: List of integer labels to be reordered
    - start: The starting value for reordering

    Returns:
    - List of reordered labels
    """
    reordered_labels = []
    next_value = start
    label_to_value = {}  # Mapping from original label to new value

    for label in labels:
        if label not in label_to_value:
            label_to_value[label] = next_value
            next_value += 1

        reordered_labels.append(label_to_value[label])

    return reordered_labels


def update_labels(clust_sample: list[dict]) -> list[dict]:
    """
    Allocate new labels according to the ranking by PERCENT
    """
    clust_result = clust_sample.copy()
    clust_result.sort(key=lambda x: (-x["PERCENT"], x["LABEL"]))
    new_label = 1
    prev_label = clust_result[0]["LABEL"]
    for cs in clust_result:
        if prev_label != cs["LABEL"]:
            new_label += 1
        prev_label = cs["LABEL"]
        cs["LABEL"] = new_label
    return clust_result



from __future__ import annotations

import gzip
import hashlib
import json
import pickle
import re
from collections.abc import Iterator
from io import BufferedReader
from pathlib import Path

import pandas as pd
import wslPath

###########################################################
# Input/Output
###########################################################


# =========================================================
# SAM/BAM
# =========================================================
def read_sam(path_of_sam: str | Path) -> Iterator[list]:
    with open(path_of_sam) as f:
        for line in f:
            yield line.strip().split("\t")


# =========================================================
# Pickle
# =========================================================


def load_pickle(file_path: Path):
    with open(file_path, "rb") as f:
        return pickle.load(f)


def save_pickle(data: object, file_path: Path) -> None:
    with open(file_path, "wb") as f:
        pickle.dump(data, f)


# =========================================================
# JSONL
# =========================================================


def read_jsonl(file_path: str | Path) -> Iterator[dict]:
    with open(file_path) as f:
        for line in f:
            yield json.loads(line)


def write_jsonl(data: list[dict] | Iterator[dict], file_path: str | Path) -> None:
    with open(file_path, "w") as f:
        for d in data:
            json_str = json.dumps(d)
            f.write(json_str + "\n")


# =========================================================
# Table (Excel, CSV)
# =========================================================


def _clean_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """
    1. Strip whitespace from column names and string values in all cells.
    2. Remove NaN records
    """
    df.columns = df.columns.str.strip()  # ← Remove spaces from column names
    df = df.map(lambda x: x.strip() if isinstance(x, str) else x)
    return df.dropna()  # ← Remove NaN records


def read_xlsx(file_path: str | Path) -> list[dict[str, str]]:
    """
    Load data from an Excel file, stripping whitespace.
    """
    df = pd.read_excel(file_path)
    return _clean_dataframe(df).to_dict(orient="records")


def read_csv(file_path: str | Path) -> list[dict[str, str]]:
    """
    Load data from a CSV file with BOM handling, stripping whitespace.
    """
    # Remove BOM if present (utf-8-sig handles BOM automatically)
    df = pd.read_csv(file_path, encoding="utf-8-sig")
    return _clean_dataframe(df).to_dict(orient="records")


def write_xlsx(data: list[dict[str, str]], file_path: str | Path) -> None:
    return pd.DataFrame(data).to_excel(file_path, index=False)


# =========================================================
# FASTA / FASTQ
# =========================================================


def is_gzip_file(path_file: str | Path) -> bool:
    """Check if a file is a GZip compressed file."""
    try:
        with Path(path_file).open("rb") as f:
            return f.read(2) == b"\x1f\x8b"
    except OSError:
        return False


def read_lines(path_fastx: str | Path) -> Iterator[str]:
    if is_gzip_file(path_fastx):
        with gzip.open(path_fastx, "rt") as f:
            yield from [line.strip() for line in f]
    else:
        with open(path_fastx) as f:
            yield from [line.strip() for line in f]


def detect_fastx_format(path_fastx: str | Path) -> str:
    fastx_iterator = read_lines(path_fastx)
    lines = [next(fastx_iterator).strip() for _ in range(4)]

    if lines[0].startswith(">"):
        return "FASTA"
    elif lines[0].startswith("@") and lines[2].startswith("+"):
        return "FASTQ"
    raise ValueError("File format is unknown. The input does not appear to be in FASTA or FASTQ format.")


def read_fastq(path_fastx: str | Path) -> Iterator[dict]:
    """Read a FASTQ file and yield each record as a dictionary."""
    iterator = read_lines(path_fastx)
    try:
        while True:
            identifier = next(iterator).strip()
            sequence = next(iterator).strip()
            separator = next(iterator).strip()
            quality = next(iterator).strip()
            yield {"identifier": identifier, "sequence": sequence, "separator": separator, "quality": quality}
    except StopIteration:
        return


def read_fasta(path_fasta: str | Path) -> Iterator[dict]:
    """Read a FASTA file (multi-line supported) and yield each record as a dictionary."""
    header = None
    seq_lines: list[str] = []

    for line in read_lines(path_fasta):
        if line.startswith(">"):
            if header:
                yield {"identifier": header, "sequence": "".join(seq_lines)}
            header = line[1:].strip()  # Remove leading ">"
            seq_lines = []
        else:
            seq_lines.append(line)

    if header:
        yield {"identifier": header, "sequence": "".join(seq_lines)}


def write_fastq(data: list[dict] | Iterator[dict], file_path: str | Path, is_gzip: bool = True) -> None:
    open_func = gzip.open if is_gzip else open
    mode = "wt" if is_gzip else "w"

    with open_func(file_path, mode) as f:
        for record in data:
            f.write(f"{record['identifier']}\n{record['sequence']}\n{record['separator']}\n{record['quality']}\n")


###########################################################
# Load batch file
###########################################################


def determine_file_type(file_path: str) -> str | None:
    """Determine if the file is an Excel or CSV file. Raise error for other types."""
    file_extension = Path(file_path).suffix
    if file_extension in [".xlsx", ".xls"]:
        return "excel"
    elif file_extension == ".csv":
        return "csv"
    else:
        raise ValueError("The provided file must be either an Excel or CSV file.")


def load_batchfile(batchfile_path: str) -> list[dict[str, str]]:
    """Load data from either an Excel or CSV file."""
    file_type = determine_file_type(batchfile_path)
    if file_type == "excel":
        return read_xlsx(batchfile_path)
    elif file_type == "csv":
        return read_csv(batchfile_path)


###########################################################
# File hander
###########################################################


def count_newlines(filepath: str | Path) -> int:
    def read_in_chunks(file: BufferedReader, chunk_size: int = 2**16) -> Iterator[bytes, None, None]:
        """Get a Iterator that reads a file in chunks and yields each chunk."""
        while True:
            chunk = file.read(chunk_size)
            if not chunk:
                break
            yield chunk

    # Open the file in binary mode and count the newline characters
    with open(filepath, "rb") as file:
        count = sum(chunk.count(b"\n") for chunk in read_in_chunks(file))

    return count


def convert_to_posix(path: str) -> str:
    if wslPath.is_windows_path(path):
        path = wslPath.to_posix(path)
    return path


def sanitize_name(file_name: Path | str) -> str:
    """
    Sanitize the file name by replacing invalid characters on Windows OS with '-'.

    The regular expression r'[<>:"/\\|?*\x00-\x1f .]' matches the following characters:

    - **Windows prohibited characters**:
        `<`, `>`, `:`, `"`, `/`, `\\`, `|`, `?`, `*`
    - **Control characters (ASCII \x00 - \x1f)**:
        Examples: `\t` (tab), `\n` (newline), `\r` (carriage return)
    - **Period `.`**:
        Matches as a literal character inside `[]`. Used to remove trailing or leading dots.
    - **Space `" "`**:
        Spaces are explicitly included (`.` is preceded by `" "` in the regex),
        so they are replaced with `-` as well.

    Example:
        >>> sanitize_name("test file.txt")
        'test-file-txt'
        >>> sanitize_name("invalid|name?.txt")
        'invalid-name-.txt'
        >>> sanitize_name(" leading space.txt ")
        'leading-space-txt'
        >>> sanitize_name("file name with spaces")
        'file-name-with-spaces'
    """
    file_name = str(file_name).strip()
    if not file_name:
        raise ValueError("Provided name is empty or consists only of whitespace")

    forbidden_chars = r'[<>:"/\\|?*\x00-\x1F .]'

    return re.sub(forbidden_chars, "-", file_name)


###########################################################
# Cache handler
###########################################################


def cache_file_hash(path_input_file: str | Path, path_output_file: Path) -> None:
    """
    Generate a hash value from the content of the provided file and cache it.

    Parameters:
    - tempdir: Path object, the temporary directory path.
    - path_input_file: Path object, the path to the fasta file.
    """
    content = Path(path_input_file).read_text()

    # Calculate the hash of the content.
    content_hash = hashlib.sha256(content.encode()).hexdigest()

    # Save the hash to the file.
    path_output_file.parent.mkdir(parents=True, exist_ok=True)
    path_output_file.write_text(content_hash)



from __future__ import annotations

import uuid
from itertools import groupby
from pathlib import Path

from DAJIN2.core.clustering.clustering import return_labels
from DAJIN2.core.clustering.label_updator import relabel_with_consective_order
from DAJIN2.core.clustering.score_handler import annotate_score, make_score
from DAJIN2.core.clustering.strand_bias_handler import count_strand_distribution
from DAJIN2.utils import fileio
from DAJIN2.utils.allele_handler import to_allele_key


def extract_labels(classif_sample, TEMPDIR, SAMPLE_NAME, CONTROL_NAME) -> list[dict[str]]:
    labels_result = []
    max_label = 1
    control_key = to_allele_key("control")

    strand_sample = count_strand_distribution(
        Path(TEMPDIR, SAMPLE_NAME, "midsv", control_key, f"{SAMPLE_NAME}_midsv.jsonl")
    )
    classif_sample.sort(key=lambda x: x["ALLELE"])
    min_cluster_size = max(5, int(len(classif_sample) * 0.5 // 100))  # 0.5% of the samples

    for allele, group in groupby(classif_sample, key=lambda x: x["ALLELE"]):
        allele_key = to_allele_key(allele)
        # Cache data to temporary files

        # For insertion/inversion allele # TODO: insertion/inversion allele names may have changed
        path_control = Path(TEMPDIR, CONTROL_NAME, "midsv", allele_key, f"{SAMPLE_NAME}_midsv.jsonl")
        if not path_control.exists():
            path_control = Path(TEMPDIR, CONTROL_NAME, "midsv", allele_key, f"{CONTROL_NAME}_midsv.jsonl")

        # Write the group to a temporary JSONL file
        unique_id = str(uuid.uuid4())
        path_sample = Path(TEMPDIR, SAMPLE_NAME, "clustering", f"tmp_{allele_key}_{unique_id}.jsonl")
        fileio.write_jsonl(data=group, file_path=path_sample)

        # Load mutation_loci and knockin_loci
        path_mutation_loci = Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", allele_key, "mutation_loci.pickle")
        path_knockin_loci = Path(TEMPDIR, SAMPLE_NAME, "knockin_loci", allele_key, "knockin.pickle")

        mutation_loci: list[set[str]] = fileio.load_pickle(path_mutation_loci)
        knockin_loci: set[int] = fileio.load_pickle(path_knockin_loci) if path_knockin_loci.exists() else set()

        # Skip clustering when the number of reads is too small or there is no mutation
        read_numbers = fileio.count_newlines(path_sample)
        is_no_mutation = all(m == set() for m in mutation_loci)
        if read_numbers < max(5, min_cluster_size) or is_no_mutation:
            labels_result.extend([max_label] * read_numbers)
            max_label += 1
            continue

        # Calculate scores to temporary files
        mutation_score: list[dict[str, float]] = make_score(path_sample, path_control, mutation_loci, knockin_loci)

        scores_sample = annotate_score(path_sample, mutation_score, mutation_loci)
        scores_control = annotate_score(path_control, mutation_score, mutation_loci, is_control=True)

        path_score_sample = Path(
            TEMPDIR,
            SAMPLE_NAME,
            "clustering",
            f"tmp_{allele_key}_score_sample_{unique_id}.jsonl",
        )
        path_score_control = Path(
            TEMPDIR,
            SAMPLE_NAME,
            "clustering",
            f"tmp_{allele_key}_score_control_{unique_id}.jsonl",
        )
        fileio.write_jsonl(data=scores_sample, file_path=path_score_sample)
        fileio.write_jsonl(data=scores_control, file_path=path_score_control)

        # Extract labels
        labels = return_labels(path_score_sample, path_score_control, path_sample, strand_sample)
        labels_result += relabel_with_consective_order(labels, start=max_label)

        # Remove temporary files
        path_sample.unlink()
        path_score_sample.unlink()
        path_score_control.unlink()

        max_label = max(labels_result) + 1

    return labels_result



from __future__ import annotations

from collections import defaultdict
from pathlib import Path

from scipy.stats import fisher_exact
from sklearn.tree import DecisionTreeClassifier

from DAJIN2.utils import fileio

"""
Nanopore sequencing results often results in strand specific mutations even though the mutation is not strand specific, thus they are considered as sequencing errors and should be removed.

This module provides functions to determine whether each allele obtained after clustering is formed due to sequencing errors caused by strand bias.

Re-allocates reads belonging to clusters with strand bias to clusters without strand bias.
"""


def count_strand_distribution(path_midsv: Path) -> dict[str, int]:
    """
    Counts the distribution of '+' and '-' strands in the given control file.
    """
    strand_count = defaultdict(int)
    for sample in fileio.read_jsonl(path_midsv):
        strand_count[sample["STRAND"]] += 1

    # Ensure '+' and '-' are present for all labels
    strand_count.setdefault("+", 0)
    strand_count.setdefault("-", 0)

    return dict(strand_count)


###############################################################################
# Handle Strand bias
# # Clusters of reads with mutations with strand bias are merged into similar clusters without strand bias
###############################################################################


def count_strand_by_label(samples: dict[str, str], labels: list[int]) -> dict[int, dict[str, int]]:
    """Count the occurrences of each strand type ('+' and '-') for each label."""
    strand_counts_by_label = defaultdict(lambda: defaultdict(int))

    for label, sample in zip(labels, samples):
        strand_counts_by_label[label][sample["STRAND"]] += 1

    # Ensure '+' and '-' are present for all labels
    for counts in strand_counts_by_label.values():
        counts.setdefault("+", 0)
        counts.setdefault("-", 0)

    return {label: dict(counts) for label, counts in strand_counts_by_label.items()}


def determine_strand_biases(
    strand_counts_by_labels: dict[int, dict[str, int]], strand_sample: dict[str, int]
) -> dict[int, bool]:
    """Determine strand biases based on positive strand counts."""
    strand_biases = {}
    for label, strand_counts in strand_counts_by_labels.items():
        # numerical bias detection
        plus_ratio = strand_counts["+"] / (strand_counts["+"] + strand_counts["-"])
        is_biased_ratio = False if 0.2 < plus_ratio < 0.8 else True
        # statistical bias detection
        table = [[strand_counts["+"], strand_counts["-"]], [strand_sample["+"], strand_sample["-"]]]
        res = fisher_exact(table, alternative="two-sided")

        strand_biases[label] = is_biased_ratio and res.pvalue < 0.01

    return strand_biases


def annotate_strand_bias_by_labels(path_sample: Path, labels: list[int], strand_sample: dict[str, int]) -> bool:
    """Determine whether there is strand bias in the samples based on the provided labels."""
    samples = fileio.read_jsonl(path_sample)
    strand_counts_by_labels = count_strand_by_label(samples, labels)
    return determine_strand_biases(strand_counts_by_labels, strand_sample)


def prepare_training_testing_sets(labels, scores, strand_biases) -> tuple[list, list, list]:
    """Prepare training and testing datasets based on strand biases."""
    train_data, train_labels, test_data = [], [], []
    for label, score in zip(labels, scores):
        if strand_biases[label]:
            test_data.append(score)
        else:
            train_data.append(score)
            train_labels.append(label)
    return train_data, train_labels, test_data


def train_decision_tree(train_data, train_labels) -> DecisionTreeClassifier:
    """Train a decision tree classifier using the provided features and labels."""
    dtree = DecisionTreeClassifier(random_state=1)
    dtree.fit(train_data, train_labels)
    return dtree


def allocate_labels(labels: list[int], strand_biases: dict[str, bool], dtree, test_data) -> list[int]:
    """Re-allocates reads belonging to clusters with strand bias to clusters without strand bias."""
    label_predictions = iter(dtree.predict(test_data))
    for i, label in enumerate(labels):
        if strand_biases[label]:
            labels[i] = next(label_predictions).tolist()
    return labels


def remove_biased_clusters(
    path_sample: Path, path_score_sample: Path, labels: list[int], strand_sample: dict[str, int]
) -> list[int]:
    """Remove clusters with strand bias by re-labeling based on decision tree predictions.
    Continue until at least one of the samples exhibits strand bias (i.e., do not calculate if all samples exhibit strand bias or, if none of the samples exhibit strand bias) and
    1000 iterations are reached, which serves as a safeguard to prevent infinite loops.
    """
    strand_biases = annotate_strand_bias_by_labels(path_sample, labels, strand_sample)

    iteration_count = 0
    labels_corrected = labels
    while len(set(strand_biases.values())) > 1 and iteration_count < 1000:
        # Re-allocation of labels of biased clusters to unbiased clusters
        scores = fileio.read_jsonl(path_score_sample)
        train_data, train_labels, test_data = prepare_training_testing_sets(labels, scores, strand_biases)
        dtree = train_decision_tree(train_data, train_labels)
        labels_corrected = allocate_labels(labels, strand_biases, dtree, test_data)

        # Re-calculate strand biases based on the corrected labels
        strand_biases = annotate_strand_bias_by_labels(path_sample, labels_corrected, strand_sample)

        iteration_count += 1

    return labels_corrected



from __future__ import annotations

import os

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"

import argparse
import importlib.metadata
import logging
import shutil
import sys
from copy import deepcopy
from itertools import groupby
from pathlib import Path

from DAJIN2 import gui
from DAJIN2.core import core
from DAJIN2.utils import config, fileio, input_validator, multiprocess, report_generator

try:
    DAJIN_VERSION = importlib.metadata.version("DAJIN2")
except importlib.metadata.PackageNotFoundError:
    DAJIN_VERSION = "counld not be retrieved"


def generate_report(name: str, logger: logging.Logger) -> None:
    report_generator.report(name)
    logger.info(f"\N{PARTY POPPER} Finished! Open {config.DAJIN_RESULTS_DIR}/{name} to see the report.")


################################################################################
# Single mode
################################################################################


def execute_single_mode(arguments: dict[str]):
    # Set logging to export log to stderr and file
    path_logfile = config.get_logfile()
    log_level = logging.DEBUG if arguments["debug"] else logging.INFO
    logger = config.set_logging(path_logfile, level=log_level)
    os.environ["DAJIN2_LOGLEVEL"] = logging.getLevelName(log_level)
    logger.info(f"\N{RUNNER} Start running DAJIN2 version {DAJIN_VERSION}")
    logger.info(f"\N{PERSONAL COMPUTER} {' '.join(sys.argv)}")

    # Remove existing output directories if they exist
    if Path("DAJIN_Results", arguments["name"]).exists():
        shutil.rmtree(Path("DAJIN_Results", arguments["name"]))
    if Path("DAJIN_Results", ".tempdir", arguments["name"]).exists():
        shutil.rmtree(Path("DAJIN_Results", ".tempdir", arguments["name"]))

    # Validate input files
    input_validator.validate_files(arguments["sample"], arguments["control"], arguments["allele"])

    # Validate genome coordinates (BED file takes precedence over genome ID)
    if arguments["genome_coordinate"]:
        input_validator.validate_bed_file(arguments["genome_coordinate"])
    elif arguments["genome"]:
        # genome ID provided - use GGGenome and UCSC lookup
        # Here, two processes are executed exceptionally: validating the server and obtaining valid server information.
        # The reason is that splitting these processes would require accessing the server twice, which would increase the server load.
        arguments |= input_validator.get_available_servers()

    # Run DAJIN2
    core.execute_control(arguments)
    core.execute_sample(arguments)

    # Finish call
    generate_report(arguments["name"], logger)
    shutil.move(path_logfile, Path("DAJIN_Results", arguments["name"]))

    # Remove temporary files
    if not arguments["debug"]:
        shutil.rmtree(Path("DAJIN_Results", ".tempdir", arguments["name"]))


################################################################################
# Batch mode
################################################################################


def validate_headers_of_batch_file(headers: set[str], filepath: str) -> None:
    """Validate the headers of a batch file."""
    required_headers = {"sample", "control", "allele", "name"}
    accepted_headers = {"sample", "control", "allele", "name", "genome", "genome_coordinate", "bed"}

    if not required_headers.issubset(headers):
        raise ValueError(
            f'{filepath} must contain "sample", "control", "allele" and "name" in the header. Currently, it contains {headers}.'
        )

    if not headers.issubset(accepted_headers):
        raise ValueError(
            f'Accepted header names of {filepath} are "sample", "control", "allele", "name", "genome", "genome_coordinate", or "bed". Currently, it contains {headers}.'
        )


def create_argument_dict(args: dict, genome_with_server_urls: dict[str, str], is_control: bool) -> dict[str, str]:
    """Create a dictionary of arguments from the given headers and group."""
    args_update = deepcopy(args)

    args_update["threads"] = 1  # Set the number of threads to 1 for batch mode

    # Assign the "sample" field depending on whether it's a control or not
    if is_control:
        args_update["sample"] = args_update["control"]
    else:
        if args_update["sample"] == args_update["control"]:
            return {}  # Return an empty dict to indicate a skipped group

    # Handle genome coordinates (BED file takes precedence over genome ID)
    bed_file = args_update.get("genome_coordinate") or args_update.get("bed")

    if bed_file:
        # Ensure genome_coordinate is set for consistency
        args_update["genome_coordinate"] = bed_file
    elif args_update.get("genome"):
        args_update |= genome_with_server_urls

    return args_update


def run_DAJIN2(
    groups: list[dict[str, str]],
    genome_with_server_urls: dict[str, str],
    is_control: bool = True,
    num_workers: int = 1,
    no_filter: bool = False,
    progress_queue=None,
) -> None:
    contents = []
    for args in groups:
        args = create_argument_dict(args, genome_with_server_urls, is_control)
        if args:  # Add args to contents only if it's not an empty dict
            args["no_filter"] = no_filter  # Add no_filter to each args dict
            contents.append(args)

    # Return a list of unique dictionaries
    contents_unique = [dict(item) for item in {frozenset(d.items()) for d in contents}]

    contents_unique.sort(key=lambda x: x["sample"])

    if is_control:
        multiprocess.run(core.execute_control, contents_unique, num_workers, progress_queue=progress_queue)
    else:
        multiprocess.run(core.execute_sample, contents_unique, num_workers, progress_queue=progress_queue)


def execute_batch_mode(arguments: dict[str]):
    path_batchfile = fileio.convert_to_posix(arguments["file"])
    progress_queue = arguments.get("progress_queue")

    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")

    records = fileio.load_batchfile(path_batchfile)

    # Validate Column of the batch file
    headers = set(records[0].keys())
    validate_headers_of_batch_file(headers, path_batchfile)

    # Validate contents and fetch genome urls
    genome_with_server_urls = {}
    records.sort(key=lambda x: x["name"])
    for name, groups in groupby(records, key=lambda x: x["name"]):
        groups = list(groups)

        # Check for duplicate alleles for the same name to prevent confusion and errors. (https://github.com/akikuno/DAJIN2/issues/146)
        if len({g["allele"] for g in groups}) != 1:
            raise ValueError(
                f"Duplicate allele found for the same name: {name}. Please use unique name for each allele."
            )

        for args in groups:
            # Validate contents in the batch file
            input_validator.validate_files(args["sample"], args["control"], args["allele"])

            # Validate BED file if provided
            if args.get("genome_coordinate"):
                input_validator.validate_file_existence(args["genome_coordinate"])

            # Validate genome and fetch urls
            if args.get("genome") and args["genome"] not in genome_with_server_urls:
                genome_with_server_urls["genome"] = args["genome"]
                genome_with_server_urls |= input_validator.get_available_servers()

    # Run DAJIN2
    for name, groups in groupby(records, key=lambda x: x["name"]):
        groups: list[dict[str, str]] = list(groups)
        # Set logging to export log to stderr and file
        config.reset_logging()
        path_logfile = config.get_logfile()
        log_level = logging.DEBUG if arguments["debug"] else logging.INFO
        logger = config.set_logging(path_logfile, level=log_level)
        os.environ["DAJIN2_LOGFILE"] = str(path_logfile)
        os.environ["DAJIN2_LOGLEVEL"] = logging.getLevelName(log_level)

        logger.info(f"\N{RUNNER} Start running DAJIN2 version {DAJIN_VERSION}")
        logger.info(f"\N{PERSONAL COMPUTER} {' '.join(sys.argv)}")
        logger.info(f"\N{LEFT-POINTING MAGNIFYING GLASS} Handling {name}")

        # Remove existing output directories if they exist
        if Path("DAJIN_Results", name).exists():
            shutil.rmtree(Path("DAJIN_Results", name))
        if Path("DAJIN_Results", ".tempdir", name).exists():
            shutil.rmtree(Path("DAJIN_Results", ".tempdir", name))

        # Run DAJIN2
        run_DAJIN2(
            groups,
            genome_with_server_urls,
            is_control=True,
            num_workers=arguments["threads"],
            no_filter=arguments["no_filter"],
            progress_queue=progress_queue,
        )
        run_DAJIN2(
            groups,
            genome_with_server_urls,
            is_control=False,
            num_workers=arguments["threads"],
            no_filter=arguments["no_filter"],
            progress_queue=progress_queue,
        )

        # Finish call
        generate_report(name, logger)
        shutil.move(path_logfile, Path("DAJIN_Results", name))
        if not arguments["debug"]:
            shutil.rmtree(Path("DAJIN_Results", ".tempdir", name))
        os.environ.pop("DAJIN2_LOGFILE", None)
        os.environ.pop("DAJIN2_LOGLEVEL", None)


def execute():
    parser = argparse.ArgumentParser()

    ###############################################################################
    # Single mode
    ###############################################################################

    parser.add_argument("-s", "--sample", type=str, help="Path to a sample directory including FASTQ file")
    parser.add_argument("-c", "--control", type=str, help="Path to a control directory including FASTQ file")
    parser.add_argument("-a", "--allele", type=str, help="Path to a FASTA file")
    parser.add_argument("-n", "--name", type=str, help="Output directory name", default="Results")
    parser.add_argument(
        "-g", "--genome", type=str, default=None, help="Reference genome ID (e.g hg38, mm39) [default: '']"
    )
    parser.add_argument(
        "-b",
        "--bed",
        type=str,
        default=None,
        dest="genome_coordinate",
        help="Path to BED6 file containing genomic coordinates [default: '']",
    )
    parser.add_argument("-t", "--threads", type=int, default=1, help="Number of threads [default: 1]")
    parser.add_argument(
        "--no-filter", action="store_true", help="Disable minor allele filtering (keep alleles <0.5%%)"
    )
    parser.add_argument("-v", "--version", action="version", version=f"DAJIN2 version {DAJIN_VERSION}")
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)

    ###############################################################################
    # Batch mode
    ###############################################################################

    def batchmode(args):
        arguments = {}
        arguments["file"] = args.file
        arguments["threads"] = input_validator.update_threads(int(args.threads))
        arguments["no_filter"] = args.no_filter
        arguments["debug"] = args.debug
        execute_batch_mode(arguments)

    subparser = parser.add_subparsers()
    parser_batch = subparser.add_parser("batch", help="DAIJN2 batch mode")
    parser_batch.add_argument("-f", "--file", required=True, type=str, help="CSV or Excel file.")
    parser_batch.add_argument("-t", "--threads", default=1, type=int, help="Number of threads [default: 1]")
    parser_batch.add_argument(
        "--no-filter", action="store_true", help="Disable minor allele filtering (keep alleles <0.5%%)"
    )
    parser_batch.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    parser_batch.set_defaults(handler=batchmode)

    ###############################################################################
    # GUI mode
    ###############################################################################

    def guimode(args):
        gui.execute()

    parser_gui = subparser.add_parser("gui", help="DAIJN2 GUI mode")
    parser_gui.set_defaults(handler=guimode)

    ###############################################################################
    # Parse arguments
    ###############################################################################

    args = parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(args)
    else:
        if args.sample is None:
            parser.error("the following arguments are required: -s/--sample")
        if args.control is None:
            parser.error("the following arguments are required: -c/--control")
        if args.allele is None:
            parser.error("the following arguments are required: -a/--allele")
        if args.name is None:
            parser.error("the following arguments are required: -n/--name")

        arguments = {}
        arguments["sample"] = args.sample
        arguments["control"] = args.control
        arguments["allele"] = args.allele
        arguments["name"] = args.name
        arguments["genome"] = args.genome
        arguments["genome_coordinate"] = args.genome_coordinate
        arguments["threads"] = input_validator.update_threads(int(args.threads))
        arguments["no_filter"] = args.no_filter
        arguments["debug"] = args.debug

        execute_single_mode(arguments)


if __name__ == "__main__":
    execute()



from __future__ import annotations

from pathlib import Path
from tempfile import NamedTemporaryFile

import midsv

from DAJIN2.core.preprocess.alignment import mapping
from DAJIN2.utils import fileio
from DAJIN2.utils.allele_handler import to_allele_key
from DAJIN2.utils.dna_handler import revcomp
from DAJIN2.utils.midsv_handler import convert_midsvs_to_sequence
from DAJIN2.utils.report_name_handler import build_report_filename

LARGE_SV_THRESHOLD = 50
VALID_DNA_BASES = frozenset("ACGTN")
INFO_FIELD_PRIORITY = ("SVTYPE", "TYPE", "SVLEN", "SEQ", "QNAME")
MAPPY_PRESET = "map-ont"


def _build_vcf_record(
    chromosome_name: str,
    position: int,
    record_id: str,
    reference_bases: str,
    alternate_bases: str,
    info_fields: dict[str, str | int],
) -> dict[str, object]:
    """Create one internal VCF record object."""
    return {
        "CHROM": chromosome_name,
        "POS": position,
        "ID": record_id,
        "REF": reference_bases,
        "ALT": alternate_bases,
        "INFO": info_fields,
    }


def _format_info(info_fields: dict[str, str | int]) -> str:
    """Serialize INFO fields in a stable order for VCF output."""
    if not info_fields:
        return "."

    ordered_pairs: list[tuple[str, str | int]] = []
    emitted_keys: set[str] = set()

    for key in INFO_FIELD_PRIORITY:
        if key in info_fields:
            ordered_pairs.append((key, info_fields[key]))
            emitted_keys.add(key)

    for key in sorted(key for key in info_fields if key not in emitted_keys):
        ordered_pairs.append((key, info_fields[key]))

    serialized_pairs: list[str] = []
    for key, value in ordered_pairs:
        if isinstance(value, str):
            value = value.replace("%", "%25")
        serialized_pairs.append(f"{key}={value}")

    return ";".join(serialized_pairs)


def _reference_base(midsv_token: str) -> str:
    """Return the reference base implied by a MIDSV token."""
    token_for_reference = midsv_token
    if midsv_token.startswith("+"):
        anchor_token = next(
            (part for part in reversed(midsv_token.split("|")) if not part.startswith("+")),
            "",
        )
        token_for_reference = anchor_token or "=N"

    if len(token_for_reference) < 2:
        return "N"

    return token_for_reference[1].upper()


def _normalize_variant_sequence(sequence: str) -> str:
    """Strip non-DNA symbols from a sequence before VCF conversion."""
    return "".join(base for base in str(sequence).upper() if base in VALID_DNA_BASES)


def _revcomp_variant_sequence(sequence: str) -> str:
    """Reverse complement a sequence after removing MIDSV operators."""
    sanitized_sequence = _normalize_variant_sequence(sequence)
    return revcomp(sanitized_sequence) if sanitized_sequence else ""


def _is_inversion_token(midsv_token: str) -> bool:
    """Detect whether a MIDSV token contains inversion bases."""
    return any(character.islower() for character in midsv_token if character.isalpha())


def _parse_insertion_token(midsv_token: str) -> tuple[str, str, str, str]:
    """Extract inserted bases and anchor context from an insertion token."""
    token_parts = midsv_token.split("|")
    inserted_sequence = "".join(part[1:].upper() for part in token_parts if part.startswith("+"))
    anchor_token = next((part for part in reversed(token_parts) if not part.startswith("+")), "")
    anchor_operation = anchor_token[0] if anchor_token else "="
    anchor_body = anchor_token[1:].upper() if len(anchor_token) > 1 else "N"
    anchor_reference_base = anchor_body[0] if anchor_body else "N"
    anchor_alternate_bases = anchor_reference_base

    if anchor_operation == "*" and len(anchor_body) >= 2:
        anchor_alternate_bases = anchor_body[1:]
    elif anchor_operation == "-":
        anchor_alternate_bases = ""

    return anchor_operation, anchor_reference_base, anchor_alternate_bases, inserted_sequence


def _is_unknown_token(midsv_token: str) -> bool:
    """Detect placeholder tokens that represent unknown bases."""
    return midsv_token.upper() in {"=N", "N"}


def _build_inversion_record(
    chromosome_name: str,
    start_position: int,
    inversion_bases: list[str],
    allele_record_id: str,
) -> dict[str, object] | None:
    """Build a symbolic inversion record from buffered inversion bases."""
    if start_position is None:
        return None

    reference_sequence = "".join(inversion_bases).upper() or "N"
    info_fields: dict[str, str | int] = {
        "SVTYPE": "INV",
        "SVLEN": len(inversion_bases),
        "SEQ": reference_sequence,
        "QNAME": allele_record_id,
    }
    return _build_vcf_record(
        chromosome_name,
        start_position,
        allele_record_id,
        reference_sequence[0],
        "<INV>",
        info_fields,
    )


def _build_unknown_run_record(
    chromosome_name: str,
    start_position: int,
    run_length: int,
    allele_record_id: str,
) -> dict[str, object] | None:
    """Represent an unknown-base run as a symbolic deletion-like record."""
    if start_position is None:
        return None

    reference_sequence = "N" * run_length
    info_fields: dict[str, str | int] = {
        "SVTYPE": "DEL",
        "TYPE": "DEL",
        "SVLEN": -run_length,
        "SEQ": reference_sequence,
        "QNAME": allele_record_id,
    }
    return _build_vcf_record(
        chromosome_name,
        start_position,
        allele_record_id,
        reference_sequence[0],
        "<DEL>",
        info_fields,
    )


def _collect_deleted_sequence(midsv_tokens: list[str], start_index: int) -> tuple[str, int]:
    """Merge consecutive deletion tokens into one deleted sequence."""
    deleted_bases = midsv_tokens[start_index][1:].upper()
    consumed_token_count = 1

    for lookahead_token in midsv_tokens[start_index + 1 :]:
        if not lookahead_token.startswith("-") or _is_inversion_token(lookahead_token):
            break
        deleted_bases += lookahead_token[1:].upper()
        consumed_token_count += 1

    return deleted_bases, consumed_token_count


def _build_insertion_alt(reference_base: str, inserted_sequence: str, large_sv_threshold: int) -> str:
    """Choose symbolic or explicit ALT notation for an insertion."""
    if len(inserted_sequence) > large_sv_threshold:
        return "<INS>"
    return reference_base + inserted_sequence


def _midsv_to_vcf_records(
    midsv_tags: list[str],
    chrom: str,
    start_offset: int,
    allele_id: str,
    large_sv_threshold: int,
) -> list[dict[str, object]]:
    """Convert MIDSV tokens into local-coordinate VCF records."""
    records: list[dict[str, object]] = []
    local_position = start_offset + 1
    left_anchor_base: str | None = None
    inversion_start_position: int | None = None
    inversion_bases: list[str] = []
    unknown_run_start_position: int | None = None
    unknown_run_length = 0

    def flush_inversion_record() -> None:
        """Emit the buffered inversion event, if present."""
        nonlocal inversion_start_position, inversion_bases
        inversion_record = _build_inversion_record(chrom, inversion_start_position, inversion_bases, allele_id)
        if inversion_record is not None:
            records.append(inversion_record)
        inversion_start_position = None
        inversion_bases = []

    def flush_unknown_run_record() -> None:
        """Emit the buffered unknown-base run, if present."""
        nonlocal unknown_run_start_position, unknown_run_length
        unknown_run_record = _build_unknown_run_record(
            chrom,
            unknown_run_start_position,
            unknown_run_length,
            allele_id,
        )
        if unknown_run_record is not None:
            records.append(unknown_run_record)
        unknown_run_start_position = None
        unknown_run_length = 0

    # Walk through MIDSV tokens from left to right while tracking the current
    # local reference position and any event that spans multiple tokens.
    token_index = 0
    while token_index < len(midsv_tags):
        current_token = midsv_tags[token_index]

        # Inversions may cover consecutive lowercase tokens, so keep buffering
        # them until a non-inversion token appears.
        if _is_inversion_token(current_token):
            flush_unknown_run_record()
            if inversion_start_position is None:
                inversion_start_position = local_position
            current_reference_base = _reference_base(current_token)
            inversion_bases.append(current_reference_base)
            left_anchor_base = current_reference_base
            local_position += 1
            token_index += 1
            continue
        flush_inversion_record()

        # Unknown-base runs are also merged into one symbolic event.
        if _is_unknown_token(current_token):
            if unknown_run_start_position is None:
                unknown_run_start_position = local_position
            unknown_run_length += 1
            left_anchor_base = _reference_base(current_token)
            local_position += 1
            token_index += 1
            continue
        flush_unknown_run_record()

        # Pure matches only advance the reference cursor and refresh the last
        # known anchor base for downstream insertions.
        if current_token.startswith("="):
            left_anchor_base = _reference_base(current_token)
            local_position += 1
            token_index += 1
            continue

        # Substitutions become one single-base VCF record at the current
        # reference position.
        if current_token.startswith("*"):
            reference_base = _reference_base(current_token)
            alternate_base = current_token[2:].upper() if len(current_token) >= 3 else reference_base
            substitution_info: dict[str, str | int] = {"TYPE": "SUB", "QNAME": allele_id}
            records.append(
                _build_vcf_record(
                    chrom,
                    local_position,
                    allele_id,
                    reference_base,
                    alternate_base,
                    substitution_info,
                )
            )
            left_anchor_base = reference_base
            local_position += 1
            token_index += 1
            continue

        # Consecutive deletions are collapsed into one VCF deletion record with
        # a computed END coordinate.
        if current_token.startswith("-"):
            deleted_sequence, consumed_token_count = _collect_deleted_sequence(midsv_tags, token_index)
            deletion_length = len(deleted_sequence)
            deletion_end_position = local_position + deletion_length - 1
            deletion_info: dict[str, str | int] = {
                "SVTYPE": "DEL",
                "TYPE": "DEL",
                "SVLEN": -deletion_length,
                "SEQ": deleted_sequence,
                "QNAME": allele_id,
                "END": deletion_end_position,
            }
            records.append(
                _build_vcf_record(
                    chrom,
                    local_position,
                    allele_id,
                    deleted_sequence[0] if deleted_sequence else "N",
                    "<DEL>",
                    deletion_info,
                )
            )
            if deleted_sequence:
                left_anchor_base = deleted_sequence[-1]
            local_position += consumed_token_count
            token_index += consumed_token_count
            continue

        # Insertions use the previous matched base as the left anchor when
        # available. If the anchor token itself encodes a substitution, emit an
        # additional substitution record at the same locus.
        if current_token.startswith("+"):
            (
                anchor_operation,
                anchor_reference_base,
                anchor_alternate_bases,
                inserted_sequence,
            ) = _parse_insertion_token(current_token)
            has_left_anchor = left_anchor_base is not None and local_position > start_offset + 1
            insertion_position = local_position - 1 if has_left_anchor else local_position
            insertion_reference_base = left_anchor_base if has_left_anchor else anchor_reference_base
            insertion_info: dict[str, str | int] = {"TYPE": "INS", "SVLEN": len(inserted_sequence), "QNAME": allele_id}
            if inserted_sequence:
                insertion_info["SEQ"] = inserted_sequence
            records.append(
                _build_vcf_record(
                    chrom,
                    insertion_position,
                    allele_id,
                    insertion_reference_base,
                    _build_insertion_alt(insertion_reference_base, inserted_sequence, large_sv_threshold),
                    insertion_info,
                )
            )
            if anchor_operation == "*" and anchor_alternate_bases and anchor_alternate_bases != anchor_reference_base:
                substitution_info: dict[str, str | int] = {"TYPE": "SUB", "QNAME": allele_id}
                records.append(
                    _build_vcf_record(
                        chrom,
                        local_position,
                        allele_id,
                        anchor_reference_base,
                        anchor_alternate_bases,
                        substitution_info,
                    )
                )
            left_anchor_base = anchor_reference_base
            local_position += 1
            token_index += 1
            continue

        # Any other token shape is treated conservatively as one consumed
        # reference base so the parser stays synchronized.
        left_anchor_base = _reference_base(current_token)
        local_position += 1
        token_index += 1

    # Flush buffered multi-token events that reach the end of the sequence.
    flush_inversion_record()
    flush_unknown_run_record()
    return records


def _should_align_to_control(allele_name: str) -> bool:
    """Return whether an allele should be re-aligned against control."""
    return allele_name.lower() != "control"


def _extract_allele_name(consensus_key: str) -> str:
    """Extract the allele name portion from a consensus header."""
    key_parts = consensus_key.split("|")
    if len(key_parts) >= 2:
        return key_parts[1]
    return consensus_key


def _load_control_sequence(tempdir: Path, sample_name: str) -> str | None:
    """Load the control FASTA sequence used for VCF alignment."""
    path_control_fasta = Path(tempdir, sample_name, "fasta", f"{to_allele_key('control')}.fasta")
    if not path_control_fasta.exists():
        return None

    for fasta_record in fileio.read_fasta(path_control_fasta):
        return fasta_record["sequence"]
    return None


def _write_temp_fasta(sequence: str, record_name: str) -> Path:
    """Write a temporary FASTA file and return its path."""
    with NamedTemporaryFile("w", delete=False, suffix=".fasta") as temporary_fasta:
        temporary_fasta.write(f">{record_name}\n")
        for start_index in range(0, len(sequence), 80):
            temporary_fasta.write(sequence[start_index : start_index + 80] + "\n")
        return Path(temporary_fasta.name)


def _filter_primary_sam_lines(sam_lines: list[str]) -> list[str]:
    """Remove supplementary alignments before MIDSV conversion."""
    filtered_lines: list[str] = []
    for sam_line in sam_lines:
        normalized_line = sam_line if sam_line.endswith("\n") else f"{sam_line}\n"
        if normalized_line.startswith("@"):
            filtered_lines.append(normalized_line)
            continue

        sam_fields = normalized_line.rstrip("\n").split("\t")
        if len(sam_fields) > 1:
            try:
                alignment_flag = int(sam_fields[1])
            except ValueError:
                alignment_flag = 0
            if alignment_flag & 2048:
                continue

        filtered_lines.append(normalized_line)

    return filtered_lines


def _align_sequence_to_control(control_sequence: str, query_sequence: str) -> list[str]:
    """Align a consensus sequence to control and return MIDSV tokens."""
    if control_sequence.upper() == query_sequence.upper():
        return [f"={base}" for base in control_sequence.upper()]

    path_reference_fasta: Path | None = None
    path_query_fasta: Path | None = None
    path_alignment_sam: Path | None = None

    try:
        path_reference_fasta = _write_temp_fasta(control_sequence, "control")
        path_query_fasta = _write_temp_fasta(query_sequence, "query")
        sam_lines = list(mapping.to_sam(path_reference_fasta, path_query_fasta, preset=MAPPY_PRESET))
        if not sam_lines:
            return []

        primary_sam_lines = _filter_primary_sam_lines(sam_lines)
        with NamedTemporaryFile("w", delete=False, suffix=".sam") as temporary_sam:
            path_alignment_sam = Path(temporary_sam.name)
            temporary_sam.writelines(primary_sam_lines)

        midsv_alignments = midsv.transform(path_sam=path_alignment_sam, qscore=False)
        if not midsv_alignments:
            return []

        return midsv_alignments[0]["MIDSV"].split(",")
    finally:
        for temporary_path in (path_reference_fasta, path_query_fasta, path_alignment_sam):
            if temporary_path is not None:
                temporary_path.unlink(missing_ok=True)


def _convert_record_on_negative_strand(
    record: dict[str, object],
    info_fields: dict[str, str | int],
    region_end: int,
) -> dict[str, object]:
    """Convert one local-coordinate record to genomic coordinates on the minus strand."""
    variant_kind = str(info_fields.get("TYPE") or info_fields.get("SVTYPE") or "")
    local_position = int(record["POS"])
    local_end_position = int(info_fields.get("END", local_position))
    converted_record = dict(record)
    sequence_in_info = str(info_fields.get("SEQ", ""))

    if variant_kind == "INS":
        genomic_position = region_end - local_position + 1
        reference_base = _revcomp_variant_sequence(str(record["REF"])) or "N"
        inserted_sequence = _revcomp_variant_sequence(sequence_in_info)
        alternate_bases = "<INS>" if str(record["ALT"]) == "<INS>" else reference_base + inserted_sequence
        if inserted_sequence:
            info_fields["SEQ"] = inserted_sequence
        converted_record.update({"POS": genomic_position, "REF": reference_base, "ALT": alternate_bases})
        return converted_record

    if variant_kind == "DEL":
        genomic_position = region_end - local_end_position + 1
        genomic_end_position = region_end - local_position + 1
        deleted_sequence = _revcomp_variant_sequence(sequence_in_info)
        reference_base = (
            deleted_sequence[0] if deleted_sequence else _revcomp_variant_sequence(str(record["REF"])) or "N"
        )
        info_fields["SEQ"] = deleted_sequence
        info_fields["END"] = genomic_end_position
        converted_record.update({"POS": genomic_position, "REF": reference_base, "ALT": "<DEL>"})
        return converted_record

    if variant_kind == "SUB":
        genomic_position = region_end - local_position + 1
        converted_record.update(
            {
                "POS": genomic_position,
                "REF": _revcomp_variant_sequence(str(record["REF"])) or "N",
                "ALT": _revcomp_variant_sequence(str(record["ALT"])) or "N",
            }
        )
        return converted_record

    if variant_kind == "INV":
        inverted_sequence = _revcomp_variant_sequence(sequence_in_info)
        genomic_position = region_end - local_end_position + 1
        info_fields["SEQ"] = inverted_sequence
        info_fields["END"] = region_end - local_position + 1
        converted_record.update(
            {
                "POS": genomic_position,
                "REF": inverted_sequence[0]
                if inverted_sequence
                else _revcomp_variant_sequence(str(record["REF"])) or "N",
                "ALT": "<INV>",
            }
        )
        return converted_record

    genomic_position = region_end - local_position + 1
    converted_record["POS"] = genomic_position
    if isinstance(record.get("REF"), str):
        converted_record["REF"] = _revcomp_variant_sequence(str(record["REF"])) or "N"
    if isinstance(record.get("ALT"), str) and not str(record["ALT"]).startswith("<"):
        converted_record["ALT"] = _revcomp_variant_sequence(str(record["ALT"])) or "N"
    return converted_record


def _convert_records_to_genomic_coordinates(
    records: list[dict[str, object]],
    genome_coordinates: dict[str, object] | None,
) -> list[dict[str, object]]:
    """Translate local VCF records into genomic coordinates."""
    if not records or genome_coordinates is None:
        return records

    chromosome_name = str(genome_coordinates.get("chrom") or "control")
    region_start = int(genome_coordinates.get("start", 1))
    region_end = int(genome_coordinates.get("end", region_start))
    strand = str(genome_coordinates.get("strand") or "+")

    genomic_records: list[dict[str, object]] = []
    for record in records:
        info_fields = dict(record.get("INFO", {})) if isinstance(record.get("INFO", {}), dict) else {}
        genomic_record = dict(record)
        genomic_record["CHROM"] = chromosome_name

        if strand != "-":
            genomic_record["POS"] = region_start + int(record["POS"]) - 1
            if "END" in info_fields:
                info_fields["END"] = region_start + int(info_fields["END"]) - 1
        else:
            genomic_record = _convert_record_on_negative_strand(genomic_record, info_fields, region_end)
            genomic_record["CHROM"] = chromosome_name

        genomic_record["INFO"] = info_fields
        genomic_records.append(genomic_record)

    genomic_records.sort(key=lambda record: (str(record["CHROM"]), int(record["POS"]), str(record["ID"])))
    return genomic_records


def _ensure_deletion_end(record: dict[str, object]) -> None:
    """Populate END for deletion records when it can be derived from SVLEN."""
    info_fields = record.get("INFO", {})
    if not isinstance(info_fields, dict) or "END" in info_fields:
        return

    variant_kind = info_fields.get("SVTYPE") or info_fields.get("TYPE")
    if variant_kind != "DEL":
        return

    sv_length = info_fields.get("SVLEN")
    try:
        deletion_length = abs(int(sv_length))
    except (TypeError, ValueError):
        return

    info_fields["END"] = int(record["POS"]) + deletion_length - 1


def _write_vcf(path_output: Path, records: list[dict[str, object]]) -> None:
    """Write internal VCF records to a VCF file."""
    sortable_records: list[dict[str, object]] = []
    for output_order, record in enumerate(records):
        record_with_order = dict(record)
        record_with_order["_order"] = output_order
        sortable_records.append(record_with_order)

    sortable_records.sort(key=lambda record: (str(record["CHROM"]), int(record["POS"]), int(record["_order"])))

    path_output.parent.mkdir(parents=True, exist_ok=True)
    with open(path_output, "w", newline="\n", encoding="utf-8") as output_handle:
        output_handle.write("##fileformat=VCFv4.3\n")
        output_handle.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n")

        for record in sortable_records:
            _ensure_deletion_end(record)
            info_fields = record.get("INFO", {})
            serialized_info = _format_info(info_fields) if isinstance(info_fields, dict) else "."
            record_id = record.get("ID", ".")
            if isinstance(record_id, str):
                record_id = record_id.replace("%", "%25")
            output_handle.write(
                f"{record['CHROM']}\t{record['POS']}\t{record_id}\t{record['REF']}\t"
                f"{record['ALT']}\t.\tPASS\t{serialized_info}\n"
            )


def _resolve_vcf_midsv_tags(
    consensus_key: str,
    consensus_midsv_tags: list[str],
    control_sequence: str | None,
) -> list[str]:
    """Choose the MIDSV tokens that should be used for VCF export."""
    allele_name = _extract_allele_name(consensus_key)
    if not _should_align_to_control(allele_name) or not control_sequence:
        return consensus_midsv_tags

    consensus_sequence = convert_midsvs_to_sequence(consensus_midsv_tags).upper()
    aligned_midsv_tags = _align_sequence_to_control(control_sequence, consensus_sequence)
    return aligned_midsv_tags or consensus_midsv_tags


def export_to_vcf(
    tempdir: Path,
    sample_name: str,
    genome_coordinates: dict | None,
    cons_midsv_tags: dict[str, list[str]],
) -> None:
    """Export consensus MIDSV calls as per-allele VCF files."""
    normalized_genome_coordinates = genome_coordinates or {}
    chromosome_name = str(normalized_genome_coordinates.get("chrom") or "control")
    control_sequence = _load_control_sequence(tempdir, sample_name)

    for consensus_key, consensus_midsv_tags in cons_midsv_tags.items():
        record_id = consensus_key.replace("|", "_")
        midsv_tags_for_vcf = _resolve_vcf_midsv_tags(consensus_key, consensus_midsv_tags, control_sequence)
        local_records = _midsv_to_vcf_records(
            midsv_tags_for_vcf,
            chromosome_name,
            0,
            record_id,
            LARGE_SV_THRESHOLD,
        )
        genomic_records = _convert_records_to_genomic_coordinates(local_records, normalized_genome_coordinates)
        output_filename = build_report_filename(consensus_key, ".vcf", sample_name=sample_name)
        path_output = Path(tempdir, "report", "VCF", sample_name, output_filename)
        _write_vcf(path_output, genomic_records)



from __future__ import annotations

import random
import uuid
from collections import defaultdict
from pathlib import Path

import pysam

from DAJIN2.core.preprocess.alignment.mapping import to_sam
from DAJIN2.utils import fileio, sam_handler
from DAJIN2.utils.allele_handler import to_allele_key
from DAJIN2.utils.report_name_handler import build_report_filename


def recalculate_sam_coordinates_to_reference(sam: list[list[str]], genome_coordinates: dict) -> list[str]:
    """Recalculate SAM genomic coordinates with the reference genome, not with the FASTA_ALLELE"""
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@")]
    for s in sam_headers:
        if s[0] != "@SQ":
            continue
        s[1] = f"SN:{genome_coordinates['chrom']}"
        s[2] = f"LN:{genome_coordinates['chrom_size']}"

    for s in sam_contents:
        s[2] = genome_coordinates["chrom"]

    if genome_coordinates["strand"] == "-":
        sam_contents = sam_handler.revcomp_sam(sam_contents, genome_coordinates["end"])
    else:
        for s in sam_contents:
            s[3] = str(int(s[3]) + genome_coordinates["start"] - 1)

    return sam_headers + sam_contents


def convert_pos_to_one_indexed(sam_lines: list[list[str]]) -> list[list[str]]:
    """Convert SAM POS from 0-indexed to 1-indexed"""

    def convert_index(line: list[str]) -> list[str]:
        if not line[0].startswith("@") and line[3] == "0":
            line[3] = "1"
        return line

    return [convert_index(line) for line in sam_lines]


def group_by_allele_name(sam_contents: list[str], result_sample: list[dict]) -> dict[list]:
    """Group alignments in the input SAM by allele name (NAME)."""
    sam_contents.sort()
    result_sample_sorted = sorted(result_sample, key=lambda x: x["QNAME"])

    qnames: set[str] = {c["QNAME"] for c in result_sample_sorted}

    sam_groups = defaultdict(list)
    idx_sam_contents = 0
    idx_result_sample = 0
    while idx_sam_contents < len(sam_contents) and idx_result_sample < len(result_sample_sorted):
        alignments_sam = sam_contents[idx_sam_contents][:-1]  # Discard CS tags to reduce file size
        alignments_clsut_sample = result_sample_sorted[idx_result_sample]
        qname_sam = alignments_sam[0]

        if qname_sam not in qnames:
            idx_sam_contents += 1
            continue

        if qname_sam == alignments_clsut_sample["QNAME"]:
            key = alignments_clsut_sample["NAME"]
            sam_groups[key].append(alignments_sam)
            idx_sam_contents += 1
        else:
            idx_result_sample += 1

    return dict(sam_groups)


###############################################################################
# igvjs
###############################################################################


def sample_unique_qnames(qnames: list[str], max_reads: int, rng: random.Random) -> set[str]:
    unique_qnames = list(dict.fromkeys(qnames))
    if len(unique_qnames) <= max_reads:
        return set(unique_qnames)
    return set(rng.sample(unique_qnames, max_reads))


def sample_sam_by_qname(sam_content: list[list[str]], max_reads: int, rng: random.Random) -> list[list[str]]:
    qnames = [sam[0] for sam in sam_content]
    sampled_qnames = sample_unique_qnames(qnames, max_reads, rng)
    return [sam for sam in sam_content if sam[0] in sampled_qnames]


###############################################################################
# Output
###############################################################################


def split_headers_and_contents(sam: list) -> tuple:
    """Split SAM headers and contents."""
    sam_headers = [s for s in sam if s[0].startswith("@")]
    sam_contents = [s for s in sam if not s[0].startswith("@")]
    return sam_headers, sam_contents


def write_sam_to_bam(sam: list[list[str]], path_bam: str | Path, threads: int = 1) -> None:
    # Format SAM
    formatted_sam = "\n".join("\t".join(s) for s in sam) + "\n"

    # Write SAM to a temporary file
    path_sam = Path(path_bam.parent, f"temp_{uuid.uuid4()}.sam")
    with open(path_sam, "w", newline="\n", encoding="utf-8") as f:
        f.write(formatted_sam)

    # Convert SAM to BAM
    pysam.sort("-@", f"{threads}", "-o", str(path_bam), str(path_sam))
    pysam.index("-@", f"{threads}", str(path_bam))
    # Remove temporary file
    path_sam.unlink()


def update_genome_coordinates(sam: list, genome_coordinates: dict | None) -> list:
    if genome_coordinates is None:
        genome_coordinates = {}

    sam_records = sam.copy()
    sam_records = sam_handler.remove_microhomology(sam_records)
    # Update @SQ headers if genome coordinates are available (from --genome or --genome-coordinate)
    if genome_coordinates.get("chrom") and genome_coordinates.get("chrom_size"):
        return recalculate_sam_coordinates_to_reference(sam_records, genome_coordinates)
    else:
        return convert_pos_to_one_indexed(sam_records)


def resolve_best_sam_path(tempdir: str | Path, name: str, allele: str = "control") -> Path:
    allele_key = to_allele_key(allele)
    path_best_sam = Path(tempdir, name, "midsv", allele_key, f"{name}_best.sam")
    if path_best_sam.exists():
        return path_best_sam
    return Path(tempdir, name, "sam", allele_key, "map-ont.sam")


def export_sequence_error_to_bam(TEMPDIR, NAME, genome_coordinates, THREADS) -> None:
    path_fastq = Path(TEMPDIR, NAME, "fastq", f"{NAME}_sequence_error.fastq.gz")
    path_fasta = Path(TEMPDIR, NAME, "fasta", f"{to_allele_key('control')}.fasta")
    preset = "map-ont"

    sam_records = [record.split("\t") for record in to_sam(path_fasta, path_fastq, preset=preset, threads=THREADS)]
    sam_updated = update_genome_coordinates(sam_records, genome_coordinates)

    # Output SAM and BAM
    path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}_sequence_error.bam")
    write_sam_to_bam(sam_updated, path_bam_output, THREADS)


def export_to_bam(TEMPDIR, NAME, genome_coordinates, THREADS, RESULT_SAMPLE=None, is_control=False) -> None:
    path_sam_input = resolve_best_sam_path(TEMPDIR, NAME, allele="control")

    sam_records = list(fileio.read_sam(path_sam_input))
    sam_updated = update_genome_coordinates(sam_records, genome_coordinates)

    rng = random.Random()
    max_reads_igvjs = 100

    # Output SAM and BAM
    path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, f"{NAME}.bam")
    write_sam_to_bam(sam_updated, path_bam_output, THREADS)

    # Prepare SAM headers and contents
    sam_headers, sam_contents = split_headers_and_contents(sam_updated)

    if is_control:
        sam_subset = sample_sam_by_qname(sam_contents, max_reads_igvjs, rng)
        path_bam_output = Path(TEMPDIR, "cache", ".igvjs", NAME, "control.bam")
        path_bam_output.parent.mkdir(parents=True, exist_ok=True)
        write_sam_to_bam(sam_headers + sam_subset, path_bam_output, THREADS)
    else:
        sam_groups = group_by_allele_name(sam_contents, RESULT_SAMPLE)
        # Output SAM and BAM
        for name, sam_content in sam_groups.items():
            # BAM
            file_name = build_report_filename(name, ".bam", sample_name=NAME)
            path_bam_output = Path(TEMPDIR, "report", "BAM", NAME, file_name)
            write_sam_to_bam(sam_headers + sam_content, path_bam_output, THREADS)
            # igvjs
            sam_subset = sample_sam_by_qname(sam_content, max_reads_igvjs, rng)
            file_name = build_report_filename(name, ".bam")
            path_bam_output = Path(TEMPDIR, "report", ".igvjs", NAME, file_name)
            write_sam_to_bam(sam_headers + sam_subset, path_bam_output, THREADS)

    # Export sequence error to BAM
    export_sequence_error_to_bam(TEMPDIR, NAME, genome_coordinates, THREADS)



from __future__ import annotations


def revcomp(sequence: str) -> str:
    """Reverse complement a DNA sequence."""
    complement = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N"}
    return "".join(complement[nt] for nt in sequence[::-1])



"""
Anomaly detection utilities for mutation extraction.
"""

from __future__ import annotations

import numpy as np
from sklearn.neural_network import MLPClassifier


def cosine_distance(x: list[float], y: list[float]) -> float:
    """Calculate cosine distance between two vectors."""
    x_arr = np.asarray(x, dtype=float)
    y_arr = np.asarray(y, dtype=float)

    if x_arr.size == 0 or y_arr.size == 0:
        return 0.0

    denominator = np.linalg.norm(x_arr) * np.linalg.norm(y_arr)
    if denominator == 0 or np.isnan(denominator):
        return 1.0

    cosine_similarity = float(np.dot(x_arr, y_arr) / (denominator + np.finfo(float).eps))
    cosine_similarity = np.clip(cosine_similarity, -1.0, 1.0)
    return 1 - cosine_similarity


def is_dissimilar_loci(values_sample, values_control, index: int, is_consensus: bool = False) -> bool:
    """Check if a locus shows dissimilar mutation patterns between sample and control."""
    # If 'sample' has more than 20% variation compared to 'control' in consensus mode,
    # unconditionally set it to 'dissimilar loci'. This is set to counteract cases where,
    # when evaluating cosine similarity during significant deletions, values exceedingly
    # close to 1 can occur even if not observed in the control
    # (e.g., control = [1,1,1,1,1], sample = [100,100,100,100,100] -> cosine similarity = 1).
    if values_sample[index] - values_control[index] > 20:
        if not is_consensus or values_sample[index] > 50:
            return True
        else:
            return False

    # Subset 10 bases around index.
    x = values_sample[index : index + 10]
    y = values_control[index : index + 10]

    x_slice = x[1:]
    y_slice = y[1:]

    distance = cosine_distance(x, y)
    distance_slice = cosine_distance(x_slice, y_slice)

    if distance <= 0.05:
        return False

    ratio_denominator = distance + distance_slice
    if ratio_denominator == 0:
        return False

    return distance / ratio_denominator > 0.9


def detect_anomalies(values_sample, values_control, threshold: float, is_consensus: bool = False) -> set[int]:
    """
    Detect anomalies and return indices of outliers.
    """
    rng = np.random.default_rng(seed=1)

    random_size = 10_000
    control_size = len(values_control)
    total_size = random_size + control_size

    randoms = rng.uniform(0, 100, random_size)
    randoms_error = np.clip(randoms + rng.uniform(0, threshold, random_size), 0, 100)
    randoms_mutation = np.clip(randoms + rng.uniform(threshold, 100, random_size), 0, 100)

    values_error = np.clip(values_control + rng.uniform(0, threshold, control_size), 0, 100)
    values_mutation = np.clip(values_control + rng.uniform(threshold, 100, control_size), 0, 100)

    matrix_error_randoms = np.array([randoms, randoms_error]).T
    matrix_error_control = np.array([values_control, values_error]).T
    matrix_error = np.concatenate([matrix_error_randoms, matrix_error_control], axis=0)

    matrix_mutation_randoms = np.array([randoms, randoms_mutation]).T
    matrix_mutation_control = np.array([values_control, values_mutation]).T
    matrix_mutation = np.concatenate([matrix_mutation_randoms, matrix_mutation_control], axis=0)

    X = np.concatenate([matrix_error, matrix_mutation], axis=0)
    y = [0] * (total_size) + [1] * (total_size)

    clf = MLPClassifier(solver="lbfgs", alpha=1e-5, hidden_layer_sizes=(5, 2), random_state=1, max_iter=1000)
    clf.fit(X, y)

    results = clf.predict(np.array([values_control, values_sample]).T)

    candidate_loci = {i for i, v in enumerate(results) if v == 1}

    return {i for i in candidate_loci if is_dissimilar_loci(values_sample, values_control, i, is_consensus)}


def extract_anomal_loci(
    indels_normalized_sample,
    indels_normalized_control,
    thresholds: dict[str, float],
    is_consensus: bool = False,
) -> dict[str, set[int]]:
    """Extract outlier loci comparing indel counts between sample and control."""
    anomal_loci = {}
    for mut in {"+", "-", "*"}:
        values_sample = indels_normalized_sample[mut]
        values_control = indels_normalized_control[mut]
        idx_outliers = detect_anomalies(values_sample, values_control, thresholds[mut], is_consensus)
        anomal_loci[mut] = idx_outliers
    return anomal_loci



from __future__ import annotations

import re

###############################################################################
# HTML Builder
###############################################################################

HTML_HEADER = """<!DOCTYPE html>
<html lang="en">
<meta charset="utf-8">

<head>
    <style>
        body {
            width: 95%;
            margin: 0 auto;
            box-sizing: border-box;
            font-family: Consolas, 'Courier New', monospace;
            color: #333;
        }

        h1 {
            padding: 0.1em 0;
            border-top: solid 3px #333;
            border-bottom: solid 3px #333;
        }

        .p_seq {
            color: #585858;
            word-wrap: break-word;
            letter-spacing: 0.15em;
        }

        .p_legend {
            word-wrap: break-word;
        }

        .p_legend span {
            margin: 5px;
        }

        .p_legend br {
            line-height: 2em;
        }

        .Ins {
            font-weight: bold;
            border: 0.1em solid;
            background-color: #ee827c;
        }

        .Del {
            font-weight: bold;
            border: 0.1em solid;
            background-color: #a0d8ef;
        }

        .Sub {
            font-weight: bold;
            border: 0.1em solid;
            background-color: #98d98e;
        }

        .Unknown {
            font-weight: bold;
            border: 0.1em solid;
            background-color: #c0c6c9;
        }

        .Inv {
            font-weight: bold;
            text-decoration: underline;
            text-decoration-color: #7F4DFF;
            text-decoration-thickness: 3px;
            text-underline-offset: auto;
        }
    </style>
</head>

<body>
"""

HTML_LEGEND = """
<p class="p_legend">
    <span class="Ins">Insertion</span>
    <span class="Del">Deletion</span>
    <span class="Sub">Substitution</span>
    <span class="Unknown">Unknown</span>
    <span class="Inv">Inversion</span>
</p>
<hr>
"""

HTML_FOOTER = """
</body>
</html>
"""


###############################################################################
# Highlight SV regions
###############################################################################


def get_inversion_range(midsv_sv_allele) -> list[tuple[int, int]]:
    inversion_range = []
    prev_inversion = False
    inversion_start = -1
    for i, tag_sv in enumerate(midsv_sv_allele):
        if tag_sv.islower() and prev_inversion is False:
            inversion_start = i
            prev_inversion = True
        elif not tag_sv.islower() and prev_inversion is True:
            inversion_range.append((inversion_start, i - 1))
            inversion_start = -1
            prev_inversion = False

        # Last tag is inversion
        if i == len(midsv_sv_allele) - 1 and prev_inversion is True:
            inversion_range.append((inversion_start, i))

    return inversion_range


def get_insertion_range(midsv_sv_allele) -> list[tuple[int, int]]:
    insertion_range = []
    idx = 0
    for i, tag_sv in enumerate(midsv_sv_allele):
        if tag_sv.startswith("+"):
            insertion_start = i + idx
            insertion_end = insertion_start + tag_sv.count("|") - 1
            insertion_range.append((insertion_start, insertion_end))
            idx += insertion_end
    return insertion_range


def add_sv_class(cons_midsv_tag, sv_range, class_name: str) -> list[str]:
    sv_range = iter(sv_range)
    highlight_sv_allele = []

    sv_start = -1
    sv_end = -1
    for i, tag_sample in enumerate(cons_midsv_tag):
        if sv_start == -1 or i == sv_end + 1:
            try:
                sv_start, sv_end = next(sv_range)
            except StopIteration:
                pass
        if i == sv_start == sv_end:
            highlight_sv_allele.append(f"<span class='{class_name}'>{tag_sample}</span>")
        elif i == sv_start:
            highlight_sv_allele.append(f"<span class='{class_name}'>{tag_sample}")
        elif i == sv_end:
            highlight_sv_allele.append(f"{tag_sample}</span>")
        else:
            highlight_sv_allele.append(tag_sample)

    return highlight_sv_allele


def append_inversion_allele(midsv_sv_allele, cons_midsv_tag) -> list[str]:
    inversion_range = get_inversion_range(midsv_sv_allele)
    return add_sv_class(cons_midsv_tag, inversion_range, "Inv")


def append_insertion_allele(midsv_sv_allele, cons_midsv_tag) -> list[str]:
    insertion_range = get_insertion_range(midsv_sv_allele)
    return add_sv_class(cons_midsv_tag, insertion_range, "Ins_Allele")


def split_html_tags(highlight_sv_allele: list[str]) -> list[str]:
    """
    Splits a list of strings by HTML tags.

    Example:
        ["<span>=A</span>", "<span>-C", "=C", "</span>"]
        -> ["<span>", "=A", "</span>", "<span>", "-C", "=C", "</span>"]
    """
    pattern = r"(<[^>]+>)"  # Regular expression to capture HTML tags
    return [part for text in highlight_sv_allele for part in re.split(pattern, text) if part]


###############################################################################
# Embed mutations to HTML
###############################################################################


def embed_mutations_to_html(highlight_sv_allele: list[str]) -> list[str]:
    idx = 0
    html_parts = ["<p class='p_seq'>"]

    while idx < len(highlight_sv_allele):
        tag_sample = highlight_sv_allele[idx]

        if tag_sample.startswith("<"):  # HTML tag
            html_parts.append(tag_sample)
            idx += 1
            continue

        # Insertion
        if tag_sample.startswith("+"):
            insertions, last_tag = tag_sample.split("|")[:-1], tag_sample.split("|")[-1]
            insertions = "".join(ins[-1] for ins in insertions)
            html_parts.append("<span class='Ins'>" + "".join(insertions) + "</span>" + last_tag[-1])
            idx += 1
            continue

        # Substitution
        if tag_sample.startswith("*"):
            substitutions = []
            while idx < len(highlight_sv_allele) and highlight_sv_allele[idx].startswith("*"):
                substitutions.append(highlight_sv_allele[idx][-1])
                idx += 1
            html_parts.append("<span class='Sub'>" + "".join(substitutions) + "</span>")
            continue

        # Deletion
        if tag_sample.startswith("-"):
            deletions = []
            while idx < len(highlight_sv_allele) and highlight_sv_allele[idx].startswith("-"):
                deletions.append(highlight_sv_allele[idx][-1])
                idx += 1
            html_parts.append("<span class='Del'>" + "".join(deletions) + "</span>")
            continue

        # Inversion: Just convert to upper case
        if tag_sample.islower() and not tag_sample.startswith("<"):  # avoid HTML span tag
            inversions = []
            while (
                idx < len(highlight_sv_allele)
                and highlight_sv_allele[idx].islower()
                and not highlight_sv_allele[idx].startswith("<")
            ):
                tag_inversion = highlight_sv_allele[idx].upper()
                if tag_inversion.startswith("*"):  # Substitution
                    inversions.append("<span class='Sub'>" + tag_inversion[-1] + "</span>")
                elif tag_inversion.startswith("-"):  # Deletion
                    inversions.append("<span class='Del'>" + tag_inversion[-1] + "</span>")
                elif tag_inversion.startswith("+"):  # Insertion
                    tag_insertion, tag_last = tag_inversion.split("|")[:-1], tag_inversion.split("|")[-1]
                    insertion = "".join(ins[-1] for ins in tag_insertion)
                    inversions.append("<span class='Ins'>" + "".join(insertion) + "</span>" + tag_last[-1])
                else:
                    inversions.append(tag_inversion[-1])
                idx += 1
            html_parts.append("".join(inversions))
            continue

        # Othres
        html_parts.append(tag_sample[-1])
        idx += 1

    html_parts.append("</p>")

    return html_parts


###############################################################################
# main
###############################################################################


def to_html(
    midsv_sv_allele: list[str], cons_midsv_tag: list[str:], allele: str, is_sv_allele: bool, description: str = ""
) -> str:
    """Output HTML string showing a sequence with mutations colored"""
    description_str = f"<h1>{description}</h1>" if description else ""
    if allele.startswith("insertion") and is_sv_allele:
        highlight_sv_allele = append_insertion_allele(midsv_sv_allele, cons_midsv_tag)
    elif allele.startswith("inversion") and is_sv_allele:
        highlight_sv_allele = append_inversion_allele(midsv_sv_allele, cons_midsv_tag)
    else:
        highlight_sv_allele = cons_midsv_tag

    highlight_sv_allele = split_html_tags(highlight_sv_allele)

    html_parts = embed_mutations_to_html(highlight_sv_allele)
    html_parts_str = "".join(html_parts)

    return "\n".join(
        [
            HTML_HEADER,
            description_str,
            HTML_LEGEND,
            html_parts_str,
            HTML_FOOTER,
        ]
    )



from __future__ import annotations

from pathlib import Path

from rapidfuzz import process
from rapidfuzz.distance import DamerauLevenshtein

from DAJIN2.utils import fileio
from DAJIN2.utils.allele_handler import to_allele_key


def extract_unique_sv(
    fasta_sv_alleles: dict[str, str], FASTA_ALLELES: dict[str, str], base_num: int = 10
) -> dict[str, str]:
    """
    Extract unique SVs (insertions/inversions) alleles if they are dissimilar to the FASTA_ALLELES input by the user.
    "Unique SV alleles" are defined as sequences that have a difference of more than 10 bases compared to the sequences in FASTA_ALLELES
    """
    fasta_sv_alleles_unique = fasta_sv_alleles.copy()

    # Remove SV alleles that are similar to the FASTA_ALLELES input by the user
    to_delete = set()
    for key, seq in fasta_sv_alleles_unique.items():
        _, distances, _ = zip(*process.extract_iter(seq, FASTA_ALLELES.values(), scorer=DamerauLevenshtein.distance))
        if any(d < base_num for d in distances):
            to_delete.add(key)

    # Remove SV alleles that are similar to each other
    for key, seq in fasta_sv_alleles_unique.items():
        if key in to_delete:
            continue
        _, distances, _ = zip(
            *process.extract_iter(seq, fasta_sv_alleles_unique.values(), scorer=DamerauLevenshtein.distance)
        )
        similar_index = {
            k if d < base_num else None for k, d in zip(fasta_sv_alleles_unique.keys(), distances) if k != key
        }
        to_delete |= similar_index

    return {k: v for k, v in fasta_sv_alleles_unique.items() if k not in to_delete}


###########################################################
# Update keys to avoid duplicating user-specified alleles
###########################################################


def _check_duplicates_of_sets(set1: set[str], set2: set[str]) -> bool:
    """if True, there are duplicates."""
    union_set = set1.union(set2)
    total_elements = len(set1) + len(set2)
    return len(union_set) != total_elements


def add_unique_allele_keys(
    fasta_sv_alleles: dict[str, str], FASTA_ALLELES: dict[str, set], key: str
) -> dict[str, str]:
    """
    Update keys to avoid duplicating user-specified alleles.
    If the allele 'insertion01' exists in FASTA_ALLELES, increment the digits.
    (insertion01 -> insertion001 -> insertion0001...)
    Also, add _DAJIN2predicted suffix to indicate that it is a DAJIN2-predicted allele.
    """
    user_defined_alleles = set(FASTA_ALLELES)
    key_duplicated_alleles = {allele for allele in user_defined_alleles if key in allele}

    if key_duplicated_alleles == set():
        return {f"{key}{(i + 1):02}_DAJIN2predicted": value for i, value in enumerate(fasta_sv_alleles.values())}

    key_candidate_alleles = set()
    num_digits = 3  # 001
    while _check_duplicates_of_sets(key_candidate_alleles, key_duplicated_alleles):
        key_candidate_alleles = {
            f"{key}{(i + 1):0{num_digits}}_DAJIN2predicted" for i, _ in enumerate(fasta_sv_alleles)
        }
        num_digits += 1

    return dict(zip(key_candidate_alleles, fasta_sv_alleles.values()))


###########################################################
# Save fasta and midsv files
###########################################################


def save_fasta(TEMPDIR: Path | str, SAMPLE_NAME: str, fasta_sv_alleles: dict[str, str]) -> None:
    Path(TEMPDIR, SAMPLE_NAME, "fasta").mkdir(parents=True, exist_ok=True)
    for header, seq in fasta_sv_alleles.items():
        allele_key = to_allele_key(header)
        Path(TEMPDIR, SAMPLE_NAME, "fasta", f"{allele_key}.fasta").write_text(f">{header}\n{seq}\n")


def save_midsv(TEMPDIR: Path | str, SAMPLE_NAME: str, midsv_sv_alleles: dict[str, list[str]]) -> None:
    Path(TEMPDIR, SAMPLE_NAME, "midsv").mkdir(parents=True, exist_ok=True)
    for header, midsv_tag in midsv_sv_alleles.items():
        allele_key = to_allele_key(header)
        fileio.write_jsonl(midsv_tag, Path(TEMPDIR, SAMPLE_NAME, "midsv", f"consensus_{allele_key}_midsv.jsonl"))



from __future__ import annotations

import hashlib
import json
from collections.abc import Iterable
from pathlib import Path


def to_allele_key(allele_name: str) -> str:
    """Return a stable MD5 hex key for an allele name used in internal paths."""
    return hashlib.md5(allele_name.encode("utf-8")).hexdigest()


def build_allele_name_map(allele_names: Iterable[str]) -> dict[str, dict[str, str]]:
    """
    Build bidirectional maps between original allele names and their MD5 keys to
    avoid 'OSError: [Errno 36] File name too long' due to long file paths.
    """
    name_to_key = {name: to_allele_key(name) for name in sorted(set(allele_names))}
    key_to_name = {key: name for name, key in name_to_key.items()}
    return {"name_to_key": name_to_key, "key_to_name": key_to_name}


def save_allele_name_map(tempdir: str | Path, allele_names: Iterable[str]) -> Path:
    """Persist the allele-name/key mapping JSON under tempdir/cache and return its path."""
    path_output = Path(tempdir, "cache", "allele_name_map.json")
    path_output.parent.mkdir(parents=True, exist_ok=True)
    path_output.write_text(json.dumps(build_allele_name_map(allele_names), indent=2), encoding="utf-8")
    return path_output



from __future__ import annotations

import hashlib

MAX_FILENAME_BYTES = 255


def _truncate_utf8(value: str, max_bytes: int) -> str:
    """Trim a string to a UTF-8 byte limit without breaking multibyte characters."""
    encoded = value.encode("utf-8")
    if len(encoded) <= max_bytes:
        return value
    trimmed = encoded[:max_bytes]
    while trimmed and (trimmed[-1] & 0xC0) == 0x80:
        trimmed = trimmed[:-1]
    return trimmed.decode("utf-8", errors="ignore")


def _compact_consensus_label(label: str) -> str:
    """Shorten consensus labels by keeping allele id and percentage when possible."""
    normalized = label.replace("|", "_")
    parts = normalized.split("_")
    if len(parts) >= 2 and parts[0].startswith("allele") and parts[-1].endswith("%"):
        return f"{parts[0]}_{parts[-1]}"
    return normalized


def build_report_filename(
    label: str,
    extension: str,
    sample_name: str | None = None,
    max_filename_bytes: int = MAX_FILENAME_BYTES,
) -> str:
    """
    Build a safe report filename with compaction and hash fallback for long names to
    avoid 'OSError: [Errno 36] File name too long'.
    Example: "sample1_allele01_{long_allele_name}_12.3%.fasta" -> "sample1_allele01_12.3%.fasta" (if under limit).
    """
    normalized = label.replace("|", "_")
    prefix = f"{sample_name}_" if sample_name else ""
    filename = f"{prefix}{normalized}{extension}"
    if len(filename.encode("utf-8")) <= max_filename_bytes:
        return filename

    compact = _compact_consensus_label(label)
    filename = f"{prefix}{compact}{extension}"
    if len(filename.encode("utf-8")) <= max_filename_bytes:
        return filename

    digest = hashlib.md5(normalized.encode("utf-8")).hexdigest()[:10]
    suffix = f"_{digest}{extension}"
    max_stem_bytes = max_filename_bytes - len(suffix.encode("utf-8"))
    stem = _truncate_utf8(f"{prefix}{compact}", max_stem_bytes)
    return f"{stem}{suffix}"



from DAJIN2.core.preprocess.alignment.mapping import generate_sam
from DAJIN2.core.preprocess.alignment.midsv_caller import generate_midsv
from DAJIN2.core.preprocess.error_correction.sequence_error_handler import (
    detect_sequence_error_reads,
    replace_midsv_without_sequence_errors,
    split_fastq_by_sequence_error,
)
from DAJIN2.core.preprocess.error_correction.strand_bias_handler import extract_sequence_errors_in_strand_biased_loci
from DAJIN2.core.preprocess.genome_coordinate.genome_fetcher import fetch_chromosome_size, fetch_coordinates
from DAJIN2.core.preprocess.infrastructure.directory_manager import (
    create_report_directories,
    create_temporal_directories,
)
from DAJIN2.core.preprocess.infrastructure.input_formatter import format_inputs
from DAJIN2.core.preprocess.mutation_processing.knockin_handler import extract_knockin_loci
from DAJIN2.core.preprocess.mutation_processing.mutation_extractor import cache_mutation_loci
from DAJIN2.core.preprocess.structural_variants.sv_detector import detect_sv_alleles



from __future__ import annotations

import csv
import json
import logging
import os
import queue
import shutil
import socket
import subprocess
import threading
import time
import webbrowser
from contextlib import closing, redirect_stderr
from pathlib import Path
from threading import Timer

from flask import Flask, Response, jsonify, render_template, request, send_file
from waitress import serve
from werkzeug.utils import secure_filename

from DAJIN2 import main
from DAJIN2.utils import config

# Global variable to store progress information
progress_queues = {}
analysis_results = {}
GUI_UPLOAD_ROOT_DIR = Path(config.DAJIN_RESULTS_DIR, ".gui_upload")
MIN_THREADS = 1
MAX_THREADS = 32
SINGLE_MODE_INPUT_TYPES = {"fastq", "fasta", "bam"}


def build_completion_message(result_path: str, is_batch: bool = False) -> str:
    scope = "batch analysis" if is_batch else "analysis"
    return f"Your {scope} results are saved in the following directory: {result_path}"


def normalize_project_name(project_name: str) -> str:
    sanitized_name = secure_filename(project_name.strip())
    if not sanitized_name:
        raise ValueError("Project name must include letters, numbers, underscores, hyphens, or dots.")
    return sanitized_name


def parse_threads(raw_value: str | None, default: int = 1) -> int:
    try:
        threads = int(raw_value) if raw_value else default
    except (TypeError, ValueError):
        return default

    return max(MIN_THREADS, min(MAX_THREADS, threads))


class ProgressLogHandler(logging.Handler):
    """Custom logging handler to capture DAJIN2 log messages and send them to progress queue."""

    def __init__(self, progress_queue, echo_to_stdout: bool = False):
        super().__init__()
        self.progress_queue = progress_queue
        self.echo_to_stdout = echo_to_stdout
        self.setLevel(logging.INFO)
        # Format to match the log file format: timestamp, level, message
        formatter = logging.Formatter("%(asctime)s, %(levelname)s, %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
        self.setFormatter(formatter)

    def emit(self, record):
        try:
            # Only capture DAJIN2 related logs to avoid spam
            if record.name.startswith("DAJIN2") or "DAJIN" in record.getMessage():
                # Format the log message
                log_message = self.format(record)
                # Send to progress queue
                self.progress_queue.put({"status": "log", "message": log_message, "timestamp": time.time()})
                if self.echo_to_stdout:
                    print(log_message, flush=True)
        except Exception:
            # Don't let logging errors break the analysis
            pass


def convert_to_windows_path(unix_path):
    """Convert Unix/WSL path to Windows path"""
    import re

    # Convert WSL mount paths (/mnt/drive/) to Windows drive letters
    wsl_pattern = r"^/mnt/([a-z])(/.*)?$"
    match = re.match(wsl_pattern, unix_path.lower())

    if match:
        drive_letter = match.group(1).upper()
        rest_of_path = match.group(2) or ""
        # Convert forward slashes to backslashes
        windows_path = f"{drive_letter}:{rest_of_path}".replace("/", "\\")
        return windows_path
    else:
        # For non-WSL paths, just convert slashes
        return unix_path.replace("/", "\\")


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def validate_file(file):
    """Validate uploaded file"""
    if not file.filename:
        raise ValueError("No file selected")

    # Check file extension
    allowed_extensions = {
        "fastq": [".fastq", ".fq", ".fastq.gz", ".fq.gz"],
        "fasta": [".fasta", ".fa", ".fasta.gz", ".fa.gz"],
        "bam": [".bam"],
        "csv": [".csv"],
        "excel": [".xlsx", ".xls"],
        "bed": [".bed"],
    }

    filename = file.filename.lower()
    for file_type, extensions in allowed_extensions.items():
        if any(filename.endswith(ext) for ext in extensions):
            return file_type

    raise ValueError(f"Unsupported file format: {file.filename}")


def collect_valid_uploaded_files(files, expected_types: set[str], field_name: str):
    """Collect uploaded files and validate their formats."""
    valid_files = []
    invalid_files = []

    for file in files:
        if not file or not file.filename:
            continue

        try:
            file_type = validate_file(file)
        except ValueError:
            invalid_files.append(Path(file.filename).name)
            continue

        if file_type not in expected_types:
            invalid_files.append(Path(file.filename).name)
            continue

        valid_files.append(file)

    if invalid_files:
        invalid_file_names = ", ".join(invalid_files[:5])
        if len(invalid_files) > 5:
            invalid_file_names += ", ..."
        raise ValueError(f"{field_name} contains unsupported files: {invalid_file_names}")

    if not valid_files:
        raise ValueError(f"{field_name} does not contain supported files.")

    return valid_files


def save_uploaded_files_to_directory(files, output_dir: Path) -> list[str]:
    """Save uploaded files as flat files to match DAJIN2 directory expectations."""
    saved_paths = []
    for file in files:
        safe_filename = secure_filename(file.filename)
        if not safe_filename:
            continue

        output_path = output_dir / safe_filename
        file.save(output_path)
        saved_paths.append(str(output_path))

    return saved_paths


app = Flask(__name__)
LOGO_PATH = Path(__file__).resolve().parent / "image" / "dajin2-logo.png"


def is_wsl_environment() -> bool:
    if os.name != "posix":
        return False

    if os.environ.get("WSL_DISTRO_NAME") or os.environ.get("WSL_INTEROP"):
        return True

    try:
        with open("/proc/sys/kernel/osrelease") as kernel_info:
            return "microsoft" in kernel_info.read().lower()
    except OSError:
        return False


def open_url_in_windows_default_browser(url: str) -> bool:
    commands = [
        ["cmd.exe", "/c", "start", "", url],
        ["powershell.exe", "-NoProfile", "-Command", f'Start-Process "{url}"'],
    ]

    for command in commands:
        try:
            subprocess.run(
                command,
                check=True,
                timeout=5,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return True
        except (FileNotFoundError, subprocess.CalledProcessError, subprocess.TimeoutExpired):
            continue

    return False


@app.route("/")
def root_page():
    return render_template("gui.html")


@app.route("/image/dajin2-logo.png")
def dajin_logo():
    return send_file(LOGO_PATH, mimetype="image/png")


@app.route("/submit-batch", methods=["POST"])
def submit_batch():
    try:
        # Validate batch file
        if "batch-file" not in request.files or not request.files.get("batch-file").filename:
            return jsonify({"error": "Please select a batch file"}), 400

        batch_file = request.files["batch-file"]
        file_type = validate_file(batch_file)

        if file_type not in ["csv", "excel"]:
            return jsonify({"error": "Batch file must be CSV or Excel format"}), 400

        # Setup directories
        import uuid

        batch_id = f"batch_{uuid.uuid4().hex[:8]}"
        TEMPDIR = Path(config.TEMP_ROOT_DIR, batch_id)
        if TEMPDIR.exists():
            shutil.rmtree(TEMPDIR)
        UPLOAD_FOLDER = Path(TEMPDIR, "upload")
        UPLOAD_FOLDER.mkdir(parents=True, exist_ok=True)

        # Save batch file
        batch_filename = secure_filename(batch_file.filename)
        batch_file_path = UPLOAD_FOLDER / batch_filename
        batch_file.save(batch_file_path)

        # Get threads and no-filter options
        threads = parse_threads(request.form.get("batch-threads"))

        no_filter = request.form.get("batch-no-filter") == "on"

        # Create progress queue for this analysis
        progress_queue = queue.Queue()
        progress_queues[batch_id] = progress_queue

        # Prepare arguments
        arguments = {
            "file": str(batch_file_path),
            "threads": threads,
            "debug": False,
            "no_filter": no_filter,
            "progress_queue": progress_queue,
        }

        # Start analysis in background thread
        def run_batch_analysis():
            try:
                # Set up log handler to capture DAJIN2 log messages
                log_handler = ProgressLogHandler(progress_queue, echo_to_stdout=True)

                # Add handler to root logger to capture ALL logs (filtered in handler)
                root_logger = logging.getLogger()
                root_logger.addHandler(log_handler)

                # Also add to DAJIN2 main logger for good measure
                dajin_logger = logging.getLogger("DAJIN2")
                dajin_logger.addHandler(log_handler)
                dajin_logger.setLevel(logging.INFO)

                # Send initial status
                progress_queue.put(
                    {"status": "log", "message": "Starting batch analysis...", "timestamp": time.time()}
                )

                # Run the actual batch analysis
                main.execute_batch_mode(arguments)

                # Store result
                analysis_results[batch_id] = {
                    "status": "completed",
                    "result_path": "DAJIN_Results",
                    "abs_result_path": str(Path("DAJIN_Results").resolve()),
                    "timestamp": time.time(),
                }

                result_directory = str(Path("DAJIN_Results").resolve())
                completion_message = build_completion_message(result_directory, is_batch=True)
                progress_queue.put(
                    {
                        "status": "completed",
                        "message": completion_message,
                        "result_path": result_directory,
                    }
                )

                # Clean up handlers
                root_logger.removeHandler(log_handler)
                dajin_logger.removeHandler(log_handler)

            except FileNotFoundError as e:
                # Clean up handlers before handling error
                try:
                    root_logger.removeHandler(log_handler)
                    dajin_logger.removeHandler(log_handler)
                except (UnboundLocalError, AttributeError):
                    pass

                # Enhanced error message for file/directory not found
                current_dir = Path.cwd().resolve()
                error_str = str(e)

                # Extract the missing path from the error message
                import re

                path_match = re.search(r"'([^']*)'", error_str)
                missing_path = path_match.group(1) if path_match else "unknown path"

                detailed_error = (
                    f"File or directory not found: '{missing_path}'\n\n"
                    f"Current working directory: {current_dir}\n\n"
                    f"The path '{missing_path}' does not exist in the current directory. "
                    f"Please check the following:\n"
                    f"1. Update the paths in your batch file to match the actual file locations\n"
                    f"2. Ensure all sample, control, and allele files are accessible from '{current_dir}'\n"
                    f"3. Use absolute paths in your batch file, or\n"
                    f"4. Change the working directory where you started DAJIN2 GUI to match your batch file paths\n\n"
                    f"Original error: {error_str}"
                )

                progress_queue.put({"status": "error", "message": detailed_error})
                analysis_results[batch_id] = {"status": "error", "error": detailed_error, "timestamp": time.time()}

            except Exception as e:
                # Clean up handlers before handling error
                try:
                    root_logger.removeHandler(log_handler)
                    dajin_logger.removeHandler(log_handler)
                except (UnboundLocalError, AttributeError):
                    pass

                error_msg = f"Batch analysis error occurred: {str(e)}"
                progress_queue.put({"status": "error", "message": error_msg})
                analysis_results[batch_id] = {"status": "error", "error": error_msg, "timestamp": time.time()}

        # Start analysis thread
        analysis_thread = threading.Thread(target=run_batch_analysis)
        analysis_thread.daemon = True
        analysis_thread.start()

        return jsonify({"status": "started", "message": "Batch analysis started", "analysis_id": batch_id})

    except ValueError as e:
        return jsonify({"error": str(e)}), 400
    except Exception as e:
        return jsonify({"error": f"Unexpected error occurred: {str(e)}"}), 500


@app.route("/submit", methods=["POST"])
def submit():
    try:
        # Get form data
        name = request.form.get("name", "")
        if not name.strip():
            return jsonify({"error": "Please enter a project name"}), 400

        name = normalize_project_name(name)

        # Validate required files
        sample_files = request.files.getlist("sample")
        if not sample_files:
            return jsonify({"error": "Please select sample files"}), 400

        control_files = request.files.getlist("control")
        if not control_files:
            return jsonify({"error": "Please select a control file"}), 400

        allele_files = request.files.getlist("allele")
        if not allele_files:
            return jsonify({"error": "Please select allele files"}), 400

        sample_files = collect_valid_uploaded_files(sample_files, SINGLE_MODE_INPUT_TYPES, "Sample directory")
        control_files = collect_valid_uploaded_files(control_files, SINGLE_MODE_INPUT_TYPES, "Control directory")
        allele_files = collect_valid_uploaded_files(allele_files, {"fasta"}, "Allele file")

        if len(allele_files) != 1:
            return jsonify({"error": "Please select exactly one allele FASTA file (.fa or .fasta)"}), 400

        # Setup directories.
        # Keep GUI uploads outside DAJIN2 runtime tempdir to avoid cleanup collisions.
        GUI_TEMPDIR = Path(GUI_UPLOAD_ROOT_DIR, name)
        if GUI_TEMPDIR.exists():
            shutil.rmtree(GUI_TEMPDIR)
        UPLOAD_FOLDER = Path(GUI_TEMPDIR, "upload")
        UPLOAD_FOLDER.mkdir(parents=True, exist_ok=True)
        app.config["UPLOAD_FOLDER"] = UPLOAD_FOLDER

        # Handle directory structure for DAJIN2
        # DAJIN2 expects directories for sample/control, files for allele

        # Process sample directory upload
        sample_dir = UPLOAD_FOLDER / "sample"
        sample_dir.mkdir(parents=True, exist_ok=True)
        save_uploaded_files_to_directory(sample_files, sample_dir)

        PATH_SAMPLE = str(sample_dir)

        # Process control directory upload
        control_dir = UPLOAD_FOLDER / "control"
        control_dir.mkdir(parents=True, exist_ok=True)
        save_uploaded_files_to_directory(control_files, control_dir)

        PATH_CONTROL = str(control_dir)

        # Upload a single allele file directly (DAJIN2 expects one path in the batch CSV).
        allele_file = allele_files[0]
        allele_name = secure_filename(Path(allele_file.filename).name)
        allele_path = UPLOAD_FOLDER / allele_name
        allele_file.save(allele_path)
        PATH_ALLELE = str(allele_path)

        # Handle BED file upload (optional)
        PATH_BED = None
        if "bed" in request.files and request.files.get("bed").filename:
            bed_file = request.files["bed"]
            bed_file = collect_valid_uploaded_files([bed_file], {"bed"}, "BED file")[0]
            bed_name = secure_filename(Path(bed_file.filename).name)
            bed_path = Path(UPLOAD_FOLDER, bed_name)
            bed_file.save(bed_path)
            PATH_BED = str(bed_path)

        genome = request.form.get("genome", "").strip()
        threads = parse_threads(request.form.get("threads"))

        # Get no-filter option
        no_filter = request.form.get("no-filter") == "on"

        # Prepare batch data
        data = []
        row = {"sample": PATH_SAMPLE, "name": name, "control": PATH_CONTROL, "allele": PATH_ALLELE}
        if genome:
            row["genome"] = genome
        if PATH_BED:
            row["bed"] = PATH_BED
        data.append(row)

        # Create batch file
        batch_file_path = Path(GUI_TEMPDIR, "upload", "batch.csv")
        batch_file_path.parent.mkdir(parents=True, exist_ok=True)

        with open(batch_file_path, mode="w", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=data[0].keys())
            writer.writeheader()
            writer.writerows(data)

        # Create progress queue for this analysis
        progress_queue = queue.Queue()
        progress_queues[name] = progress_queue

        # Prepare arguments
        arguments = {
            "file": str(batch_file_path),
            "threads": threads,
            "debug": False,
            "no_filter": no_filter,
            "progress_queue": progress_queue,
        }

        # Start analysis in background thread
        def run_analysis():
            try:
                # Set up log handler to capture DAJIN2 log messages
                log_handler = ProgressLogHandler(progress_queue, echo_to_stdout=True)

                # Add handler to root logger to capture ALL logs (filtered in handler)
                root_logger = logging.getLogger()
                root_logger.addHandler(log_handler)

                # Also add to DAJIN2 main logger for good measure
                dajin_logger = logging.getLogger("DAJIN2")
                dajin_logger.addHandler(log_handler)
                dajin_logger.setLevel(logging.INFO)

                # Send initial status
                progress_queue.put({"status": "log", "message": "Starting analysis...", "timestamp": time.time()})

                # Run the actual analysis
                main.execute_batch_mode(arguments)

                # Store result
                result_path = f"DAJIN_Results/{name}"
                abs_result_path = str(Path(result_path).resolve())
                analysis_results[name] = {
                    "status": "completed",
                    "result_path": result_path,
                    "abs_result_path": abs_result_path,
                    "timestamp": time.time(),
                }

                completion_message = build_completion_message(abs_result_path)
                progress_queue.put(
                    {"status": "completed", "message": completion_message, "result_path": abs_result_path}
                )

                # Clean up handlers
                root_logger.removeHandler(log_handler)
                dajin_logger.removeHandler(log_handler)

            except FileNotFoundError as e:
                # Clean up handlers before handling error
                try:
                    root_logger.removeHandler(log_handler)
                    dajin_logger.removeHandler(log_handler)
                except (UnboundLocalError, AttributeError):
                    pass

                # Enhanced error message for file/directory not found
                current_dir = Path.cwd().resolve()
                error_str = str(e)

                # Extract the missing path from the error message
                import re

                path_match = re.search(r"'([^']*)'", error_str)
                missing_path = path_match.group(1) if path_match else "unknown path"

                detailed_error = (
                    f"File or directory not found: '{missing_path}'\n\n"
                    f"Current working directory: {current_dir}\n\n"
                    f"The path '{missing_path}' does not exist in the current directory. "
                    f"Please check the following:\n"
                    f"1. Ensure all uploaded files were saved correctly\n"
                    f"2. Verify that sample and control directories contain the expected files\n"
                    f"3. Check that allele and BED files (if provided) are accessible\n"
                    f"4. Try uploading the files again if the issue persists\n\n"
                    f"Original error: {error_str}"
                )

                progress_queue.put({"status": "error", "message": detailed_error})
                analysis_results[name] = {"status": "error", "error": detailed_error, "timestamp": time.time()}

            except Exception as e:
                # Clean up handlers before handling error
                try:
                    root_logger.removeHandler(log_handler)
                    dajin_logger.removeHandler(log_handler)
                except (UnboundLocalError, AttributeError):
                    pass

                error_msg = f"Analysis error occurred: {str(e)}"
                progress_queue.put({"status": "error", "message": error_msg})
                analysis_results[name] = {"status": "error", "error": error_msg, "timestamp": time.time()}

        # Start analysis thread
        analysis_thread = threading.Thread(target=run_analysis)
        analysis_thread.daemon = True
        analysis_thread.start()

        return jsonify({"status": "started", "message": "Analysis started", "analysis_id": name})

    except ValueError as e:
        return jsonify({"error": str(e)}), 400
    except Exception as e:
        return jsonify({"error": f"Unexpected error occurred: {str(e)}"}), 500


@app.route("/progress/<analysis_id>")
def get_progress(analysis_id):
    """Server-Sent Events endpoint for progress updates"""

    def generate():
        if analysis_id not in progress_queues:
            yield f"data: {json.dumps({'status': 'error', 'message': 'Analysis not found'})}\n\n"
            return

        progress_queue = progress_queues[analysis_id]

        while True:
            try:
                # Wait for progress update with timeout
                update = progress_queue.get(timeout=1)
                yield f"data: {json.dumps(update)}\n\n"

                # If analysis is completed or failed, stop sending updates
                if update.get("status") in ["completed", "error"]:
                    break

            except queue.Empty:
                # Send heartbeat to keep connection alive
                yield f"data: {json.dumps({'status': 'heartbeat'})}\n\n"

                # Check if analysis result is available (in case we missed the queue message)
                if analysis_id in analysis_results:
                    result = analysis_results[analysis_id]
                    if result["status"] == "completed":
                        abs_path = result.get("abs_result_path", result["result_path"])
                        is_batch = analysis_id.startswith("batch_")
                        completion_message = build_completion_message(abs_path, is_batch=is_batch)
                        yield f"data: {json.dumps({'status': 'completed', 'message': completion_message, 'result_path': abs_path})}\n\n"
                        break
                    elif result["status"] == "error":
                        yield f"data: {json.dumps({'status': 'error', 'message': result['error']})}\n\n"
                        break
            except Exception as e:
                yield f"data: {json.dumps({'status': 'error', 'message': f'Progress monitoring error: {str(e)}'})}\n\n"
                break

    return Response(generate(), mimetype="text/event-stream")


@app.route("/status/<analysis_id>")
def get_status(analysis_id):
    """Get current status of analysis"""
    if analysis_id in analysis_results:
        return jsonify(analysis_results[analysis_id])
    elif analysis_id in progress_queues:
        return jsonify({"status": "running", "message": "Analysis running..."})
    else:
        return jsonify({"status": "not_found", "message": "Analysis not found"}), 404


@app.route("/open-folder/<analysis_id>")
def open_result_folder(analysis_id):
    """Open result folder in system file manager"""
    import os
    import platform
    import subprocess

    try:
        if analysis_id not in analysis_results:
            return jsonify({"error": "Analysis not found"}), 404

        result = analysis_results[analysis_id]
        if result["status"] != "completed":
            return jsonify({"error": "Analysis not completed"}), 400

        result_path = result["result_path"]
        abs_path = str(Path(result_path).resolve())

        # Check if directory exists
        if not Path(abs_path).exists():
            return jsonify({"error": f"Result directory not found: {abs_path}"}), 404

        # Open folder in system file manager
        system = platform.system()

        # Convert path for Windows display
        windows_path = convert_to_windows_path(abs_path) if system == "Windows" else abs_path

        try:
            if system == "Windows":
                # Try multiple methods for Windows
                try:
                    # Method 1: Use explorer with Windows path
                    subprocess.run(["explorer", windows_path], check=True, timeout=5)
                except (subprocess.CalledProcessError, subprocess.TimeoutExpired):
                    try:
                        # Method 2: Use os.startfile with Windows path
                        os.startfile(windows_path)
                    except OSError:
                        # Method 3: Try with Unix path as fallback
                        os.startfile(abs_path)

            elif system == "Darwin":  # macOS
                subprocess.run(["open", abs_path], check=True, timeout=5)
            elif system == "Linux":
                subprocess.run(["xdg-open", abs_path], check=True, timeout=5)
            else:
                return jsonify(
                    {
                        "error": f"Unsupported operating system: {system}",
                        "path": abs_path,
                        "windows_path": windows_path,
                    }
                ), 400

        except Exception as open_error:
            # Log the specific error for debugging
            error_details = f"System: {system}, Path: {abs_path}, Error: {str(open_error)}"
            return jsonify(
                {
                    "error": f"Could not open folder: {str(open_error)}",
                    "debug_info": error_details,
                    "path": abs_path,
                    "windows_path": windows_path,
                    "suggestion": "Please manually navigate to the path shown above",
                }
            ), 500

        return jsonify(
            {
                "success": True,
                "message": "Folder opened successfully",
                "path": abs_path,
                "windows_path": windows_path,
                "system": system,
            }
        )

    except Exception as e:
        return jsonify({"error": f"Failed to open folder: {str(e)}"}), 500


def open_browser(PORT):
    url = f"http://localhost:{PORT}/"

    if is_wsl_environment():
        if not open_url_in_windows_default_browser(url):
            print(f"Could not open the host browser automatically. Access '{url}' manually.")
        return

    webbrowser.open_new(url)


def execute():
    PORT = find_free_port()
    print(f"Assess 'http://localhost:{PORT}/' if a browser does not automatically open.")
    Timer(1, open_browser, [PORT]).start()
    with open(os.devnull, "w") as devnull, redirect_stderr(devnull):
        serve(app, host="0.0.0.0", port=PORT)



from __future__ import annotations

from collections import defaultdict
from itertools import groupby
from pathlib import Path

from DAJIN2.core.consensus.consensus_formatter import merge_duplicated_cons_others, merge_duplicated_cons_sequences
from DAJIN2.core.consensus.sv_annotator import annotate_insertions_within_deletion, annotate_sv_allele
from DAJIN2.utils import fileio
from DAJIN2.utils.allele_handler import to_allele_key
from DAJIN2.utils.config import ConsensusKey
from DAJIN2.utils.midsv_handler import call_sequence

###########################################################
# call position weight matrix (cons_pergentage)
###########################################################


def convert_to_percentage(
    midsv_tags: list[list[str]], mutation_loci: list[set[str]], sequence: str
) -> list[dict[str, float]]:
    """
    Convert sequences and mutations into percentages, annotating sequence errors as "SEQERROR".
    """
    # Transpose the midsv_tags to facilitate per-position processing
    midsv_tags_transposed = [list(cs) for cs in zip(*midsv_tags)]
    coverage = len(midsv_tags_transposed[0])

    cons_percentage = []
    for cs_transposed, mut_loci, nucleotide in zip(midsv_tags_transposed, mutation_loci, sequence):
        count_cs = defaultdict(float)
        for cs in cs_transposed:
            operator = cs[0]
            if operator in {"+", "-", "*"}:
                if operator not in mut_loci:
                    cs = f"={nucleotide.upper()}"
            count_cs[cs] += 1 / coverage * 100
        cons_percentage.append(dict(count_cs))

    return cons_percentage


def remove_all_n(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    for c in cons_percentage:
        if len(c) == 1 and any(key in c for key in {"=N", "=n"}):
            continue
        for key in {"=N", "=n"}:
            c.pop(key, None)

    return cons_percentage


def adjust_to_100_percent(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    adjusted_percentages = []

    for percentage_dict in cons_percentage:
        total = sum(percentage_dict.values())
        scaling_factor = 100 / total

        adjusted_dict = {key: value * scaling_factor for key, value in percentage_dict.items()}
        adjusted_percentages.append(adjusted_dict)

    return adjusted_percentages


def call_percentage(
    midsv_tags: list[list[str]], mutation_loci: list[set[str]], sequence: str
) -> list[dict[str, float]]:
    """Call position weight matrix in different loci. Non-different loci are annotated as "Match"."""
    cons_percentage = convert_to_percentage(midsv_tags, mutation_loci, sequence)
    cons_percentage = remove_all_n(cons_percentage)
    return adjust_to_100_percent(cons_percentage)


###########################################################
# main
###########################################################


def call_consensus(
    tempdir: Path, sample_name: str, fasta_alleles: dict[str, str], clust_downsampled: list[dict]
) -> tuple[dict[list], dict[str], dict[str]]:
    cons_percentages = {}
    cons_sequences = {}
    cons_midsv_tags = {}

    clust_downsampled.sort(key=lambda x: [x["ALLELE"], x["LABEL"]])
    for (allele, label), group in groupby(clust_downsampled, key=lambda x: [x["ALLELE"], x["LABEL"]]):
        clust = list(group)
        allele_key = to_allele_key(allele)
        sequence = fasta_alleles[allele]

        path_consensus = Path(tempdir, sample_name, "consensus", allele_key, str(label))
        cons_mutation_loci = fileio.load_pickle(Path(path_consensus, "mutation_loci.pickle"))

        midsv_tags = [cs["MIDSV"].split(",") for cs in clust]
        cons_percentage = call_percentage(midsv_tags, cons_mutation_loci, sequence)

        cons_midsv_tag = [max(cons, key=cons.get) for cons in cons_percentage]

        path_sv_midsv_tag = Path(tempdir, sample_name, "midsv", f"consensus_{allele_key}_midsv.jsonl")
        if path_sv_midsv_tag.exists():
            sv_midsv_tag = list(fileio.read_jsonl(path_sv_midsv_tag))
            cons_midsv_tag = annotate_sv_allele(cons_midsv_tag, sv_midsv_tag)
            if allele.startswith("deletion"):
                cons_midsv_tag = annotate_insertions_within_deletion(cons_midsv_tag, sv_midsv_tag)

        percent = clust[0]["PERCENT"]
        key = ConsensusKey(allele, label, percent)
        cons_percentages[key] = cons_percentage
        cons_sequences[key] = call_sequence(cons_percentage)
        cons_midsv_tags[key] = cons_midsv_tag

    # Merge duplicate consensus sequences and percentages
    cons_sequences, label_before_to_after = merge_duplicated_cons_sequences(cons_sequences)
    cons_percentages = merge_duplicated_cons_others(cons_percentages, cons_sequences, label_before_to_after)
    cons_midsv_tags = merge_duplicated_cons_others(cons_midsv_tags, cons_sequences, label_before_to_after)

    return cons_percentages, cons_sequences, cons_midsv_tags, label_before_to_after



from __future__ import annotations

from itertools import groupby
from pathlib import Path

from DAJIN2.core.classification.allele_merger import merge_minor_alleles
from DAJIN2.utils import fileio
from DAJIN2.utils.allele_handler import to_allele_key


def calc_match(midsv_tags: str) -> int:
    """
    Calculate match score from MIDSV tags.
    1. Perfect match: 0
    2. Mismatch (insertion, deletion, inversion): -1 per event
    """
    mismatch_score = 0

    mismatch_score += midsv_tags.count("+")  # insertion
    mismatch_score += midsv_tags.count("-")  # deletion
    mismatch_score += midsv_tags.count("*")  # substitution
    mismatch_score += sum(tags.islower() for tags in midsv_tags)  # inversion

    return -mismatch_score


def score_allele(path_midsv: Path, allele: str) -> list[dict]:
    midsv_sample = fileio.read_jsonl(path_midsv)
    scored_alleles = []
    for dict_midsv in midsv_sample:
        score = calc_match(dict_midsv["MIDSV"])
        dict_midsv.update({"SCORE": score, "ALLELE": allele})
        scored_alleles.append(dict_midsv)

    return scored_alleles


def extract_alleles_with_max_score(score_of_each_alleles: list[dict]) -> list[dict]:
    alleles_with_max_score = []
    score_of_each_alleles.sort(key=lambda x: x["QNAME"])
    for _, group in groupby(score_of_each_alleles, key=lambda x: x["QNAME"]):
        max_read = max(group, key=lambda x: x["SCORE"])
        del max_read["SCORE"]
        alleles_with_max_score.append(max_read)

    return alleles_with_max_score


##########################################################
# main
##########################################################


def classify_alleles(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str, no_filter: bool = False) -> list[dict]:
    score_of_each_alleles = []
    for allele in FASTA_ALLELES:
        path_midsv = Path(TEMPDIR, SAMPLE_NAME, "midsv", to_allele_key(allele), f"{SAMPLE_NAME}_midsv.jsonl")
        score_of_each_alleles.extend(score_allele(path_midsv, allele))

    if no_filter:
        # Skip filtering when --no-filter is used
        score_of_each_alleles_merged = score_of_each_alleles
    else:
        # Apply existing filtering logic
        score_of_each_alleles_merged = merge_minor_alleles(score_of_each_alleles, threshold=5)

    return extract_alleles_with_max_score(score_of_each_alleles_merged)



from __future__ import annotations

from collections.abc import Iterator
from pathlib import Path

import cstag
import mappy

from DAJIN2.utils import dna_handler


def to_sam(
    path_reference_fasta: Path,
    path_query_fastx: Path,
    preset: str = "map-ont",
    threads: int = 1,
    options: dict = None,
    cslong: bool = True,
) -> Iterator[str]:
    """Align sequences using mappy and Convert PAF to SAM.

    Args:
        path_reference_fasta (Path): Path of reference fasta.
        path_query_fastx (Path): Path of query fasta/fastq.
        preset (str, optional): Alignment preset. Defaults to "map-ont".
        threads (int, optional): Number of threads to use. Defaults to 1.
        cslong (bool, optional): Use long formatted CS tag if True. Defaults to True.

    Yields:
        str: SAM formatted alignment.
    """
    if options is None:
        options = {}

    path_reference_fasta = str(path_reference_fasta)
    path_query_fastx = str(path_query_fastx)

    SAM = [f"@SQ\tSN:{n}\tLN:{len(s)}" for n, s, _ in mappy.fastx_read(path_reference_fasta)]

    ref = mappy.Aligner(path_reference_fasta, preset=preset, n_threads=threads, **options)
    if not ref:
        raise ValueError(f"Failed to load {path_reference_fasta}")

    for QUERY_NAME, QUERY_SEQ, QUERY_QUAL in mappy.fastx_read(path_query_fastx):
        for hit in ref.map(QUERY_SEQ, cs=True):
            query_seq = QUERY_SEQ.upper()
            query_qual = QUERY_QUAL

            # Skip multi-mapping reads
            if hit.mapq == 0:
                continue

            # Report flag
            if hit.is_primary:
                flag = 0 if hit.strand == 1 else 16
            else:
                flag = 2048 if hit.strand == 1 else 2064

            # Handle reverse complement for negative strand
            if hit.strand == -1:
                query_seq = dna_handler.revcomp(query_seq)
                if query_qual:
                    query_qual = query_qual[::-1]

            # Append softclips to CIGAR
            cigar = hit.cigar_str
            if hit.q_st > 0:
                softclip = f"{hit.q_st}S"
                cigar = softclip + cigar if hit.strand == 1 else cigar + softclip
            if len(query_seq) - hit.q_en > 0:
                softclip = f"{len(query_seq) - hit.q_en}S"
                cigar = cigar + softclip if hit.strand == 1 else softclip + cigar

            # Convert to CS tag's long format
            if cslong:
                cs = cstag.lengthen(hit.cs, cigar, query_seq, prefix=True)

            # Summarize
            alignment = [
                QUERY_NAME,
                str(flag),
                hit.ctg,
                str(hit.r_st + 1),
                str(hit.mapq),
                cigar,
                "*",
                "0",
                "0",
                query_seq,
                "*" if query_qual is None else query_qual,
                cs,
            ]

            SAM.append("\t".join(alignment))

    yield from SAM


########################################################################
# main
########################################################################


def generate_sam(
    ARGS, paths_fasta: list[str], mappy_options: dict = None, is_control: bool = False, is_sv: bool = False
) -> None:
    if mappy_options is None:
        mappy_options = {}

    if is_control:
        path_fastq = Path(ARGS.tempdir, ARGS.control_name, "fastq", f"{ARGS.control_name}.fastq.gz")
        name = ARGS.control_name
    else:
        path_fastq = Path(ARGS.tempdir, ARGS.sample_name, "fastq", f"{ARGS.sample_name}.fastq.gz")
        name = ARGS.sample_name

    for path_fasta in paths_fasta:
        name_fasta = Path(path_fasta).stem
        presets = ["map-ont", "map-hifi", "sr"]

        path_sam_directory = Path(ARGS.tempdir, name, "sam", name_fasta)
        path_sam_directory.mkdir(parents=True, exist_ok=True)

        for preset in presets:
            sam = to_sam(path_fasta, path_fastq, preset=preset, threads=ARGS.threads, options=mappy_options)

            if is_control and is_sv:
                path_sam_file = Path(path_sam_directory, f"{ARGS.sample_name}_{preset}.sam")
            else:
                path_sam_file = Path(path_sam_directory, f"{preset}.sam")

            path_sam_file.write_text("\n".join(sam))
